"""
Creates an SQLite database that summarizes the key metrics from a collection of
Darshan logs.  The database is constructed in a way that facilitates the
determination of how much I/O is performed to different file systems.

Schemata::

    CREATE TABLE summaries (
        log_id INTEGER,
        fs_id INTEGER,

        bytes_read INTEGER,
        bytes_written INTEGER,
        reads INTEGER,
        writes INTEGER,
        file_not_aligned INTEGER,
        consec_reads INTEGER
        consec_writes INTEGER,

        mmaps INTEGER,
        opens INTEGER,
        seeks INTEGER,
        stats INTEGER,
        fdsyncs INTEGER,
        fsyncs INTEGER,

        seq_reads INTEGER,
        seq_writes INTEGER,
        rw_switches INTEGER,

        f_close_start_timestamp REAL,
        f_close_end_timestamp REAL,
        f_open_start_timestamp REAL,
        f_open_end_timestamp REAL,

        f_read_start_timestamp REAL,
        f_read_end_timestamp REAL,

        f_write_start_timestamp REAL,
        f_write_end_timestamp REAL,

        FOREIGN KEY (fs_id) REFERENCES mounts (fs_id),
        FOREIGN KEY (log_id) REFERENCES headers (log_id),
        UNIQUE(log_id, fs_id)
    );

    CREATE TABLE mounts (
        fs_id INTEGER PRIMARY KEY,
        mountpt CHAR,
        fsname CHAR
    );

    CREATE TABLE headers (
        log_id INTEGER PRIMARY KEY,
        filename CHAR UNIQUE,
        end_time INTEGER,
        exe CHAR,
        exename CHAR,
        jobid CHAR,
        nprocs INTEGER,
        start_time INTEGER,
        uid INTEGER,
        username CHAR,
        log_version CHAR,
        walltime INTEGER
    );
"""

import os
import re
import time
import sqlite3
import operator
import functools
import argparse
import warnings
import multiprocessing
import tokio.connectors.darshan

INTEGER_COUNTERS = {
    'bytes_read': operator.add,
    'bytes_written': operator.add,
    'reads': operator.add,
    'writes': operator.add,
    'file_not_aligned': operator.add,
    'consec_reads': operator.add,
    'consec_writes': operator.add,
    'mmaps': operator.add,
    'opens': operator.add,
    'seeks': operator.add,
    'stats': operator.add,
    'fdsyncs': operator.add,
    'fsyncs': operator.add,
    'flushes': operator.add,
    'seq_reads': operator.add,
    'seq_writes': operator.add,
    'rw_switches': operator.add,
    'posix_files': operator.add,
    'stdio_files': operator.add,
}

REAL_COUNTERS = {
    'f_close_start_timestamp': min,
    'f_close_end_timestamp': max,
    'f_open_start_timestamp': min,
    'f_open_end_timestamp': max,
    'f_read_start_timestamp': min,
    'f_read_end_timestamp': max,
    'f_write_start_timestamp': min,
    'f_write_end_timestamp': max,
}

SUMMARY_COUNTERS = list(INTEGER_COUNTERS.keys()) + list(REAL_COUNTERS.keys())

HEADER_COUNTERS = [
    'end_time',
    'jobid',
    'nprocs',
    'start_time',
    'uid',
    'version',
    'walltime',
]

MOUNTS_TABLE = "mounts"
HEADERS_TABLE = "headers"
SUMMARIES_TABLE = "summaries"

VERBOSITY = 0
QUIET = False

# precompile regular expressions
MOUNT_TO_FSNAME = {}

def init_mount_to_fsname():
    """Initialize regexes to map mount points to file system names
    """
    global MOUNT_TO_FSNAME
    for rex_str, fsname in tokio.config.CONFIG.get('mount_to_fsname', {}).items():
        MOUNT_TO_FSNAME[re.compile(rex_str)] = fsname

def get_file_mount(filename, mount_list):
    """Return the mount point in which a file is located

    Args:
        filename (str): Fully equalified path to a file or directory
        mount_list (list of str): List of mount points

    Returns:
        tuple of (str, str) or None: The member of mount_list in which filename
            lives; first string is the mount point, and the second is the
            logical file system name.  Returns None if filename does not match
            any mounts
    """
    # always want the most specific mount path to match first
    sorted_mount_list = sorted(mount_list, key=len, reverse=True)

    for mount in sorted_mount_list:
        if filename.startswith(mount):
            logical = mount
            for mount_rex, fsname in MOUNT_TO_FSNAME.items():
                match = mount_rex.match(mount)
                if match:
                    logical = fsname
            return (mount, logical)
        if (filename == "<STDOUT>" or filename == "<STDERR>") and mount == "UNKNOWN":
            return mount, mount
    return None;

def summarize_by_fs(darshan_log, max_mb=0.0):
    """Generates summary scalar values for a Darshan log

    Args:
        darshan_log (str): Path to a Darshan log file
        max_mb (float): Skip logs of size larger than this value

    Returns:
        dict: Contains three keys (summaries, mounts, and headers) whose values
            are dicts of key-value pairs corresponding to scalar summary values
            from the POSIX module which are reduced over all files sharing a
            common mount point.
    """
    if max_mb > 0.0 and (os.path.getsize(darshan_log) / 1024.0 / 1024.0) > max_mb:
        errmsg = "Skipping %s due to size (%d MiB)" % (darshan_log, (os.path.getsize(darshan_log) / 1024 / 1024))
        warnings.warn(errmsg)
        return {}

    try:
        darshan_data = tokio.connectors.darshan.Darshan(darshan_log, silent_errors=True)
        if QUIET:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                darshan_data.darshan_parser_base()
        else:
            darshan_data.darshan_parser_base()
    except:
        if not QUIET:
            errmsg = "Unable to open or parse %s" % darshan_log
            warnings.warn(errmsg)
        return {}

    if not darshan_data:
        if not QUIET:
            errmsg = "Unable to open or parse %s" % darshan_log
            warnings.warn(errmsg)
        return {}

    module_records = {
        'posix': darshan_data.get('counters', {}).get('posix', {}),
        'stdio': darshan_data.get('counters', {}).get('stdio', {}),
    }
    if not module_records['posix'] and not module_records['stdio']:
        if not QUIET:
            errmsg = "No counters found in %s" % darshan_log
            warnings.warn(errmsg)
        return {}

    mount_list = list(darshan_data.get('mounts', {}).keys())
    if not mount_list:
        if not QUIET:
            errmsg = "No mount table found in %s" % darshan_log
            warnings.warn(errmsg)
        return {}

    # hack in UNKNOWN for the stdio module since it does not appear in the mount table
    mount_list += ["UNKNOWN"]

    # Populate the summaries data
    reduced_counters = {}
    for mount in mount_list:
        reduced_counters[mount] = {}
        for counter in INTEGER_COUNTERS:
            reduced_counters[mount][counter] = 0
        for counter in REAL_COUNTERS:
            reduced_counters[mount][counter] = 0.0
        reduced_counters[mount]['filename'] = os.path.basename(darshan_data.log_file)

    # Reduce each counter according to its mount point
    logical_mount_names = {}
    for module, log_counters in module_records.items():
        # record_file is the full path to a file that the application manipulated
        for record_file in log_counters:
            # skip file records resident on unknown file systems
            mount = get_file_mount(record_file, mount_list)
            if mount is None:
                continue
            mount, logical = mount
            logical_mount_names[mount] = logical

            # only iterate over values; keys are MPI ranks that we don't use
            for counters in log_counters[record_file].values():
                # Now loop over the counters in our schema and add them if
                # they exist in the Darshan log's compendium.
                for counter, reduction in INTEGER_COUNTERS.items():
                    logged_val = counters.get(counter.upper())
                    if logged_val is not None:
                        reduced_counters[mount][counter] = reduction(reduced_counters[mount][counter], logged_val)
                    elif (counter == 'posix_files' and module == 'posix') \
                    or (counter == 'stdio_files' and module == 'stdio'):
                        reduced_counters[mount][counter] += 1
                    # else:
                        # if counter doesn't exist in Darshan log and isn't one
                        # of our special counters, skip it do nothing--don't
                        # attempt to reduce with an implicit zero value, as this
                        # can screw up reductions that use min/max!

                for counter, reduction in REAL_COUNTERS.items():
                    logged_val = counters.get(counter.upper())
                    if logged_val is not None:
                        reduced_counters[mount][counter] = reduction(reduced_counters[mount][counter], logged_val)

    # Populate the mounts data and remove all mount points that were not used
    mountpts = {}
    for key in list(reduced_counters.keys()):
        # check_val works only when all the keys are positive counters
        check_val = sum([reduced_counters[key][x] for x in ('bytes_read', 'bytes_written', 'reads', 'writes', 'opens', 'stats')])
        if check_val == 0:
            reduced_counters.pop(key, None)
        else:
            mountpts[key] = logical_mount_names.get(key, key)

    # Populate the headers data
    header = {}
    counters = darshan_data.get('header')
    header['filename'] = os.path.basename(darshan_data.log_file)
    for counter in HEADER_COUNTERS:
        header[counter] = counters.get(counter)

    # some Darshan logs at NERSC have no exename for some reason
    header['exe'] = counters.get('exe')
    if header['exe']:
        header['exe'] = header['exe'][0]
        header['exename'] = os.path.basename(header['exe'])
    else:
        header['exe'] = None
        header['exename'] = None

    # username is resolved here so that it can be indexed without having to mess around
    header['username'] = darshan_data.filename_metadata.get('username')

    return {
        'summaries': reduced_counters,
        'headers': header,
        'mounts': mountpts
    }

def create_mount_table(conn):
    """Creates the mount table
    """
    cursor = conn.cursor()

    query = """CREATE TABLE IF NOT EXISTS %s (
        fs_id INTEGER PRIMARY KEY,
        mountpt CHAR UNIQUE,
        fsname CHAR
    )
    """ % MOUNTS_TABLE
    vprint(query, 3)
    cursor.execute(query)

    cursor.close()
    conn.commit()

def update_mount_table(conn, mount_points):
    """Adds new mount points to the mount table
    """
    cursor = conn.cursor()

    # We force the addition of UNKNOWN so stdio's stdout/stderr counters are
    # still included
    for mount_point in mount_points.items():
        vprint("INSERT OR IGNORE INTO %s (mountpt, fsname) VALUES (?, ?)" % MOUNTS_TABLE, 4)
        vprint("Parameters: %s" % str(mount_point), 4)

    cursor.executemany("INSERT OR IGNORE INTO %s (mountpt, fsname) VALUES (?, ?)" % MOUNTS_TABLE, [(x, mount_points[x]) for x in mount_points])

    cursor.close()
    conn.commit()

def create_headers_table(conn):
    """Creates the headers table
    """
    cursor = conn.cursor()
    query = """CREATE TABLE IF NOT EXISTS %s (
        log_id INTEGER PRIMARY KEY,
        filename CHAR UNIQUE,
        end_time INTEGER,
        exe CHAR,
        exename CHAR,
        jobid CHAR,
        nprocs INTEGER,
        start_time INTEGER,
        uid INTEGER,
        username CHAR,
        version CHAR,
        walltime INTEGER
    )
    """ % HEADERS_TABLE
    vprint(query, 3)
    cursor.execute(query)
    cursor.close()
    conn.commit()

def update_headers_table(conn, header_data):
    """Adds new header data to the headers table
    """
    cursor = conn.cursor()

    header_counters = ["filename", "exe", "username", "exename"] + HEADER_COUNTERS

    # "INSERT OR IGNORE INTO" below allows duplicate summaries rows to be added.
    # This is generally harmless as long as you always INNER JOIN summaries to
    # headers when calculating reduced metrics over the whole summaries table,
    # but not all users may know to do this.
    #
    # In practice the following will only trigger a sqlite3.IntegrityError if
    # the database was corrupted in such a way that the headers table lost some
    # values but the summaries table did not.  The way the database is populated
    # here makes that impossible, but someone may hack the database on their own
    # and cause problems.
    query = "INSERT INTO %s (" % HEADERS_TABLE
    query += ", ".join(header_counters)
    query += ") VALUES (" + ",".join(["?"] * len(header_counters))
    query += ")"

    for header_datum in header_data:
        vprint(query, 4)
        vprint("Parameters: %s" % str(tuple([header_datum[x] for x in header_counters])), 4)
#       cursor.execute(query, tuple([header_datum[x] for x in header_counters])) # easier debugging
    cursor.executemany(query, [tuple([header_datum[x] for x in header_counters]) for header_datum in header_data])

    cursor.close()
    conn.commit()

def create_summaries_table(conn):
    """Creates the summaries table
    """
    cursor = conn.cursor()
    query = """CREATE TABLE IF NOT EXISTS %s (
        log_id INTEGER,
        fs_id INTEGER,
    """ % SUMMARIES_TABLE

    query += " INTEGER,\n    ".join(INTEGER_COUNTERS.keys()) + " INTEGER, \n    "
    query += " REAL,\n    ".join(REAL_COUNTERS.keys()) + " REAL,\n    "

    query += """
        FOREIGN KEY (fs_id) REFERENCES mounts (fs_id),
        FOREIGN KEY (log_id) REFERENCES headers (log_id),
        UNIQUE(log_id, fs_id)
    )
    """
    vprint(query, 3)
    cursor.execute(query)
    cursor.close()
    conn.commit()

def update_summaries_table(conn, summary_data):
    """Adds new summary counters to the summaries table
    """
    cursor = conn.cursor()

    base_query = "INSERT INTO %s (" % SUMMARIES_TABLE
    base_query += "\n  log_id,\n  fs_id,\n  " + ",\n  ".join(SUMMARY_COUNTERS) + ") VALUES (\n"
    for per_fs_counters in summary_data:
        for mountpt, summary_datum in per_fs_counters.items():
            query = base_query + "  (SELECT log_id from headers where filename = '%s'),\n" % summary_datum['filename']
            query += "  (SELECT fs_id from mounts where mountpt = '%s'),\n  " % mountpt
            query += ", ".join(["?"] * len(SUMMARY_COUNTERS))
            query += ")"
            vprint(query, 4)
            vprint("Parameters: %s" % str(tuple([summary_datum[x] for x in SUMMARY_COUNTERS])), 4)
            cursor.execute(query, tuple([summary_datum[x] for x in SUMMARY_COUNTERS]))

    cursor.close()
    conn.commit()

def get_existing_logs(conn):
    """Returns list of log files already indexed in db

    Scans the summaries table for existing entries and returns the file names
    corresponding to those entries.  We don't worry about summary rows that
    don't correspond to existing header entries because the schema prevents
    this.  Similarly, each log's summaries are committed as a single transaction
    so we can assume that if a log file has _any_ rows represented in the
    summaries table, it has been fully processed and does not need to be
    updated.

    Args:
        conn (sqlite3.Connection): Connection to database containing existing
            logs

    Returns:
        list of str: Basenames of Darshan log files represnted in the database
    """
    cursor = conn.cursor()
    # test table existence first
    cursor.execute("SELECT name FROM sqlite_master WHERE type = 'table' AND (name = ? OR name = ?)",
                   (SUMMARIES_TABLE, HEADERS_TABLE))
    if len(cursor.fetchall()) < 2:
        return []

    # retrieve all existing logs
    cursor.execute("""SELECT DISTINCT h.filename FROM %s AS s
                      INNER JOIN %s AS h ON s.log_id = h.log_id""" % (SUMMARIES_TABLE, HEADERS_TABLE))

    results = cursor.fetchall()
    return [x[0] for x in results]

def process_log_list(conn, log_list):
    """Expand and filter the list of logs to process

    Takes log_list as input by user and returns a list of Darshan logs that
    should be added to the index database.  It does the following:

    1. Expands log_list from a single-element list pointing to a directory [of
       logs] into a list of log files
    2. Returns the subset of Darshan logs which do not already appear in the
       given database.

    Relies on the logic of get_existing_logs() to determine whether a log
    appears in a database or not.  If a database is somehow created where the
    summaries table is fully populated but the headers table is not, this will
    still return log files corresponding to the missing headers and
    potentially result in duplicate summaries entries that have no matching
    header.

    Args:
        conn (sqlite3.Connection): Database containing log data
        log_list (list of str): List of paths to Darshan logs or a single-element
            list to a directory

    Returns:
        list of str: Subset of log_list that contains only those Darshan logs
            that are not already represented in the database referenced by conn.
    """
    # Filter out logs that were already processed.  NOTE: get_existing_logs()
    # returns basename; exclude_list is path/basename
    exclude_list = get_existing_logs(conn)
    vprint("%d log files already found in database" % len(exclude_list), 1)

    # If only one argument is passed in but it's a directory, enumerate all the
    # files in that directory.  This is a compromise for cases where the CLI
    # length limit prevents all darshan logs from being passed via argv, but
    # prevents every CLI arg from having to be checked to determine if it's a
    # darshan log or directory
    num_excluded = 0
    new_log_list = []
    if len(log_list) == 1 and os.path.isdir(log_list[0]):
        for filename in os.listdir(log_list[0]):
            filename = os.path.join(log_list[0], filename)
            if os.path.isfile(filename) and os.path.basename(filename) not in exclude_list:
                new_log_list.append(filename)
            else:
                num_excluded += 1
    else:
        for filename in log_list:
            if os.path.basename(filename) not in exclude_list:
                new_log_list.append(filename)
            else:
                num_excluded += 1

    vprint("Adding %d new logs" % len(new_log_list), 1)
    vprint("Excluding %d existing logs" % num_excluded, 1)

    return new_log_list

def index_darshanlogs(log_list, output_file, threads=1, max_mb=0.0):
    """Calculate the sum bytes read/written

    Given a list of input files, parse each as a Darshan log in parallel to
    create a list of scalar summary values correspond to each log and insert
    these into an SQLite database.

    Current implementation parses all logs and stores their index values in
    memory before beginning the database insert process.  This can be
    memory-intensive if processing many millions of logs at once but avoids
    thread contention on the SQLite database.

    Args:
        log_list (list of str): Paths to Darshan logs to be processed
        output_file (str): Path to a SQLite database file to populate
        threads (int): Number of subprocesses to spawn for Darshan log parsing
        max_mb (float): Skip logs of size larger than this value

    Returns:
        dict: Reduced data along different reduction dimensions
    """

    conn = sqlite3.connect(output_file)

    init_mount_to_fsname()

    t_start = time.time()
    new_log_list = process_log_list(conn, log_list)
    vprint("Built log list in %.1f seconds" % (time.time() - t_start), 2)

    # Analyze the remaining logs in parallel
    t_start = time.time()
    log_records = []
    mount_points = {}
    for result in multiprocessing.Pool(threads).imap_unordered(functools.partial(summarize_by_fs, max_mb=max_mb), new_log_list):
        if result:
            log_records.append(result)
            mount_points.update(result['mounts'])
    vprint("Ingested %d logs in %.1f seconds" % (len(log_records), time.time() - t_start), 2)

    # Create tables and indices
    t_start = time.time()
    create_mount_table(conn)
    create_headers_table(conn)
    create_summaries_table(conn)
    vprint("Initialized tables in %.1f seconds" % (time.time() - t_start), 2)

    # Insert new data that was collected in parallel
    t_start = time.time()
    update_mount_table(conn, mount_points)
    vprint("Updated mounts table in %.1f seconds" % (time.time() - t_start), 2)
    t_start = time.time()
    update_headers_table(conn, [x['headers'] for x in log_records])
    vprint("Updated headers table in %.1f seconds" % (time.time() - t_start), 2)
    t_start = time.time()
    update_summaries_table(conn, [x['summaries'] for x in log_records])
    vprint("Updated summaries table in %.1f seconds" % (time.time() - t_start), 2)

    conn.close()
    vprint("Updated %s" % output_file, 1)

    return

def vprint(string, level):
    """Print a message if verbosity is enabled

    Args:
        string (str): Message to print
        level (int): Minimum verbosity level required to print
    """

    if VERBOSITY >= level:
        print(string)

def main(argv=None):
    """Entry point for the CLI interface
    """
    global VERBOSITY
    global QUIET


    parser = argparse.ArgumentParser()
    parser.add_argument("darshanlogs", nargs="+", type=str, help="Darshan logs to process")
    parser.add_argument('-t', '--threads', default=1, type=int,
                        help="Number of concurrent processes (default: 1)")
    parser.add_argument('-o', '--output', type=str, default='darshanlogs.db', help="Name of output file (default: darshanlogs.db)")
    parser.add_argument('-m', '--max-mb', type=int, default=0, help="Maximum log file size to consider (default: 0 (disabled))")
    parser.add_argument('-v', '--verbose', action='count', default=0, help="Verbosity level (default: none)")
    parser.add_argument('-q', '--quiet', action='store_true', help="Suppress warnings for invalid Darshan logs")
    args = parser.parse_args(argv)

    VERBOSITY = args.verbose
    QUIET = args.quiet

    index_darshanlogs(log_list=args.darshanlogs,
                      threads=args.threads,
                      max_mb=args.max_mb,
                      output_file=args.output)



"""
Process the Darshan daily summary generated by either summarize_darshanlogs
or index_darshanlogs tools and generate a scoreboard of top sources of I/O based
on user, file system, and/or application.
"""

import re
import sys
import json
import sqlite3
import argparse
import collections

import tokio.config

BASE_QUERY = """
SELECT
    SUM(s.bytes_read) AS readbytes,
    SUM(s.bytes_written) AS writebytes,
    COUNT(DISTINCT h.filename) AS jobcount,
FROM
    summaries AS s
INNER JOIN
    headers AS h ON h.log_id = s.log_id,
    mounts AS m ON m.fs_id = s.fs_id
ORDER BY (readbytes+writebytes) DESC
"""

QUERY_PARAMS = collections.OrderedDict()
QUERY_PARAMS['per_user'] = {'col': 'h.username'}
QUERY_PARAMS['per_fs'] = {
    'col': 'm.fsname, m.mountpt',
    'group': 'm.fsname',
}
QUERY_PARAMS['per_exe'] = {'col': 'h.exename'}
QUERY_PARAMS['per_user_exe_fs'] = {
    'col': 'h.username || "|" || h.exename || "|" || m.fsname AS tuple',
    'group': 'tuple',
}

VERBOSITY = 0

def query_index_db(db_filenames,
                   limit_fs=None, limit_user=None, limit_exe=None,
                   exclude_fs=None, exclude_user=None, exclude_exe=None,
                   max_results=None):
    """Reduce Darshan log index by fs, user, and/or exe
    """

    where = []
    where0 = []
    if limit_fs:
        where0 = ["(m.mountpt LIKE '%s' OR m.fsname LIKE '%s')" % (limit, limit) for limit in limit_fs]
        where.append("(" + " OR ".join(where0) + ")")
    if exclude_fs:
        where0 = ["(m.mountpt NOT LIKE '%s' AND m.fsname NOT LIKE '%s')" % (limit, limit) for limit in exclude_fs]
        where.append("(" + " AND ".join(where0) + ")")

    if limit_user:
        where0 = ["h.username LIKE '%s'" % limit for limit in limit_user]
        where.append("(" + " OR ".join(where0) + ")")
    if exclude_user:
        where0 = ["h.username NOT LIKE '%s'" % limit for limit in exclude_user]
        where.append("(" + " AND ".join(where0) + ")")

    if limit_exe:
        where0 = ["h.exename LIKE '%s'" % limit for limit in limit_exe]
        where.append("(" + " OR ".join(where0) + ")")
    if exclude_exe:
        where0 = ["h.exename NOT LIKE '%s'" % limit for limit in exclude_exe]
        where.append("(" + " AND ".join(where0) + ")")

    results = collections.OrderedDict()

    for db_filename in db_filenames:
        conn = sqlite3.connect(db_filename)
        cursor = conn.cursor()

        for category, config in QUERY_PARAMS.items():

            query = BASE_QUERY

            # insert the column to group by
            query = query.replace("FROM", "    %s\nFROM" % config['col'])
            query = query.replace("ORDER", "GROUP BY %s\nORDER" % config.get('group', config['col']))

            # insert filter qualifiers
            if where:
                query = query.replace("GROUP",
                                      "WHERE\n    " + "\n    AND ".join(where) + "\nGROUP")

            # insert max number of return items
            if max_results:
                query += "\nLIMIT %d" % max_results

            vprint(query, 1)

            cursor.execute(query)
            if category not in results:
                results[category] = []
            results[category] += cursor.fetchall()

        cursor.close()
        conn.close()

    return results

def print_top(categorized_data, max_show=10):
    """
    Print the biggest I/O {users, exes, file systems}
    """
    names = {
        'per_fs': "File Systems",
        'per_user': "Users",
        'per_exe': "Applications",
        'per_user_exe_fs': "User/App/FS",
    }

    categories = 0
    for category, rankings in categorized_data.items():
        print_buffer = ""
        name = names.get(category, category)
        if categories > 0:
            print_buffer += "\n"
        print_buffer += "%2s  %40s %10s %10s %8s\n" % ('#', name, 'Read(GiB)', 'Write(GiB)', '# Jobs')
        print_buffer += '=' * 75 + "\n"
        displayed = 0
        for winner in sorted(rankings, key=lambda x: x[0] + x[1], reverse=True):
            winner_str = winner[3]
            if not winner_str:
                # Darshan logs without POSIX/STDIO counters and without
                # filename-encoded metadata can result in null strings for
                # winner[0]
                winner_str = "UNKNOWN|UNKNOWN|UNKNOWN"

            if '|' in winner_str:
                winner_str = winner_str.replace('|', ', ')

            if len(winner_str) > 40:
#               winner_str = "..." + winner_str[-37:]
                winner_str = winner_str[:19] + "..." + winner_str[-18:]

            displayed += 1
            if displayed > max_show:
                break
            print_buffer += "%2d. %40.40s %10.1f %10.1f %8d\n" % (displayed,
                                                                  winner_str,
                                                                  winner[0] / 2.0**30,
                                                                  winner[1] / 2.0**30,
                                                                  winner[2])
        if displayed > 0:
            sys.stdout.write(print_buffer)

        categories += 1

def vprint(string, level):
    """Print a message if verbosity is enabled

    Args:
        string (str): Message to print
        level (int): Minimum verbosity level required to print
    """

    if VERBOSITY >= level:
        print(string)

def main(argv=None):
    """Entry point for the CLI interface
    """
    global VERBOSITY

    parser = argparse.ArgumentParser()
    parser.add_argument("indexfile", type=str, nargs='+',
                        help="path to index database created by index_darshanlogs")
    parser.add_argument("--json", action='store_true',
                        help="output in json format")
    parser.add_argument("--max-show", type=int, default=10,
                        help="show top N users, apps, file systems")
    group_fs = parser.add_mutually_exclusive_group()
    group_fs.add_argument("--limit-fs", type=str, default=None,
                          help="only process data targeting this file system.  MUST be a fully qualified path to the mount point or injected SQL")
    group_fs.add_argument("--exclude-fs", type=str, default=None,
                          help="exclude data targeting this file system.  MUST be a fully qualified path to the mount point or injected SQL")
    group_user = parser.add_mutually_exclusive_group()
    group_user.add_argument("--limit-user", type=str, default=None,
                            help="only process logs generated by this user")
    group_user.add_argument("--exclude-user", type=str, default=None,
                            help="exclude logs generated by this user")
    group_exe = parser.add_mutually_exclusive_group()
    group_exe.add_argument("--limit-exe", type=str, default=None,
                           help="only process logs generated by this binary")
    group_exe.add_argument("--exclude-exe", type=str, default=None,
                           help="exclude logs generated by this binary")
    parser.add_argument('-v', '--verbose', action='count', default=0, help="Verbosity level (default: none)")

    args = parser.parse_args(argv)
    VERBOSITY = args.verbose

    kwargs = {
        'limit_user': args.limit_user.split(',') if args.limit_user else [],
        'limit_fs': args.limit_fs.split(',') if args.limit_fs else [],
        'limit_exe': args.limit_exe.split(',') if args.limit_exe else [],
        'exclude_user': args.exclude_user.split(',') if args.exclude_user else [],
        'exclude_fs': args.exclude_fs.split(',') if args.exclude_fs else [],
        'exclude_exe': args.exclude_exe.split(',') if args.exclude_exe else [],
    }

    results = query_index_db(args.indexfile, **kwargs)

    if args.json:
        print(json.dumps(results, indent=4, sort_keys=True))
    else:
        print_top(results, max_show=args.max_show)



"""
Retrieve the contents of an LMT database and cache it locally.
"""

import sys
import datetime
import argparse
import warnings
import tokio.debug
import tokio.timeseries
import tokio.connectors.lmtdb
import tokio.connectors.hdf5
from tokio.common import isstr

DATE_FMT = "%Y-%m-%dT%H:%M:%S"
DATE_FMT_PRINT = "YYYY-MM-DDTHH:MM:SS"

SCHEMA_VERSION = "1"

class DatasetDict(dict):
    """A dictionary containing TimeSeries objects

    Contains the TimeSeries objects being populated from an LMT database.
    Implemented as a class so that a single object can store all of the
    TimeSeries objects that are generated by multiple method calls.
    """
    def __init__(self, query_start, query_end, timestep, sort_hex=True, *args, **kwargs):
        super(DatasetDict, self).__init__(*args, **kwargs)
        self.query_start = query_start
        self.query_end = query_end
        self.timestep = timestep
        self.sort_hex = sort_hex

        self.config = {
            'datatargets/readbytes': {
                "units": "bytes",
                "delta": True,
                "column": "READ_BYTES",
            },
            'datatargets/writebytes': {
                "units": "bytes",
                "delta": True,
                "column": "WRITE_BYTES",
            },
            'fullness/bytes': {
                "units": "KiB",
                "delta": False,
                "column": "KBYTES_USED",
            },
            'fullness/bytestotal': {
                "units": "KiB",
                "delta": False,
            },
            'fullness/inodes': {
                "units": "inodes",
                "delta": False,
                "column": "INODES_USED",
            },
            'fullness/inodestotal': {
                "units": "inodes",
                "delta": False,
            },
            'dataservers/cpuload': {
                "units": "%",
                "delta": False,
                "column": "PCT_CPU",
            },
            'dataservers/memused': {
                "units": "%",
                "delta": False,
                "column": "PCT_MEMORY",
            },
            'mdservers/cpuload': {
                "units": "%",
                "delta": False,
                "column": "PCT_CPU",
            },
            'mdtargets/opens': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/closes': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/mknods': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/links': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/unlinks': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/mkdirs': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/rmdirs': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/renames': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/getxattrs': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/statfss': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/setattrs': {
                "units": "ops",
                "delta": True,
            },
            'mdtargets/getattrs': {
                "units": "ops",
                "delta": True,
            },
        }

        # Initialize raw datasets - extend query by one extra timestep so we can calculate deltas
        self.schema = tokio.connectors.hdf5.SCHEMA.get(SCHEMA_VERSION)
        if self.schema is None:
            raise KeyError("Schema version %d is not known by connectors.hdf5" % SCHEMA_VERSION)

        # We need to query past the intended query range in order to calculate
        # the differences between each row its following row
        self.query_end_plusplus = self.query_end + datetime.timedelta(seconds=timestep)

    def init_datasets(self, dataset_names, columns):
        """Populate empty datasets within self

        Creates and attachs TimeSeries objects to self based on a given column list

        Args:
            dataset_names (list of str): keys corresponding to self.config
                defining which datasets are being initialized
            columns (list of str): column names to use in the TimeSeries
                datasets being created
        """
        for dataset_name in dataset_names:
            hdf5_dataset_name = self.schema.get(dataset_name)
            if hdf5_dataset_name is None:
                warnings.warn("Skipping %s (not in schema)" % dataset_name)
            else:
                self[dataset_name] = tokio.timeseries.TimeSeries(dataset_name=hdf5_dataset_name,
                                                                 start=self.query_start,
                                                                 end=self.query_end_plusplus,
                                                                 timestep=self.timestep,
                                                                 num_columns=len(columns),
                                                                 column_names=columns,
                                                                 sort_hex=self.sort_hex)

    def finalize(self):
        """Convert datasets to deltas where necessary and tack on metadata

        Perform a few finishing actions to all datasets contained in self after
        they have been populated.  Such actions are configured entirely in
        self.config and require no external input.
        """
        self.convert_deltas(list(self.config.keys()))
        self.set_timeseries_metadata(list(self.config.keys()))

    def convert_deltas(self, dataset_names):
        """Convert datasets from absolute values to values per timestep

        Given a list of dataset names, determine if they need to be converted
        from monotonically increasing counters to counts per timestep, and
        convert those that do.  For those that don't, trim off the final row
        since it is not needed to calculate the difference between rows.

        Args:
            dataset_names (list of str): keys corresponding to self.config for
                the datasets to be converted/corrected
        """
        for dataset_name in dataset_names:
            if dataset_name in self:
                if self.config[dataset_name].get('delta'):
                    # Convert some datasets from absolute byte counts to deltas
                    self[dataset_name].convert_to_deltas()
                else:
                    # Trim off the last row from the non-delta datasets to compensate
                    # for initial over-sizing of the time range by an extra timestamp
                    self[dataset_name].trim_rows(1)

    def set_timeseries_metadata(self, dataset_names):
        """Set metadata constants (version, units, etc) on datasets and groups

        Args:
            dataset_names (list of str): keys corresponding to self.config for
                the datasets whose metadata should be set
        """
        for dataset_name in dataset_names:
            if dataset_name in self:
                self[dataset_name].dataset_metadata.update({
                    'version': SCHEMA_VERSION,
                    'units': self.config[dataset_name]['units']
                })
                self[dataset_name].group_metadata.update({'source': 'lmt'})

    def archive_mds_data(self, lmtdb):
        """Extract and encode data from LMT's MDS_DATA table

        Queries the LMT database, interprets resulting rows, and populates a
        dictionary of TimeSeries objects with those values.

        Args:
            lmtdb (LmtDb): database object
        """

        dataset_names = [
            'mdservers/cpuload',
        ]

        self.init_datasets(dataset_names, lmtdb.mds_names)

        # Now query the MDS_DATA table to get byte counts over the query time range
        results, columns = lmtdb.get_mds_data(self.query_start, self.query_end_plusplus)


        # Index the columns to speed up insertion of data
        col_map = {}
        try:
            for db_col in ['TIMESTAMP', 'MDS_ID', 'PCT_CPU']:
                col_map[db_col] = columns.index(db_col)
        except ValueError:
            raise ValueError("LMT database schema does not match expectation")

        # Loop through all the results of the timeseries query
        for row in results:
            if isstr(row[col_map['TIMESTAMP']]):
                # SQLite stores timestamps as a unicode string
                timestamp = datetime.datetime.strptime(row[col_map['TIMESTAMP']],
                                                       "%Y-%m-%d %H:%M:%S")
            else:
                # MySQL timestamps are automatically converted to datetime.datetime
                timestamp = row[col_map['TIMESTAMP']]
            target_name = lmtdb.mds_id_map[row[col_map['MDS_ID']]]
            for dataset_name in dataset_names:
                target_dbcol = self.config[dataset_name].get('column')
                # target_dbcol=PCT_CPU, target_name=snx11025n022
                if target_dbcol is not None:
                    self[dataset_name].insert_element(
                        timestamp,
                        target_name,
                        row[col_map[target_dbcol]])
                else:
                    errmsg = "%s in self.config but missing 'column' setting" % dataset_name
                    raise KeyError(errmsg)

    def archive_mds_ops_data(self, lmtdb):
        """Extract and encode data from LMT's MDS_OPS_DATA table

        Queries the LMT database, interprets resulting rows, and populates a
        dictionary of TimeSeries objects with those values.  Avoids JOINing the
        MDS_VARIABLE_INFO table and instead uses an internal mapping of
        OPERATION_IDs to demultiplex the data in MDS_OPS_DATA into different
        HDF5 datasets.

        Args:
            lmtdb (LmtDb): database object
        """

        # mapping between OPERATION_INFO.OPERATION_NAME to HDF5 dataset names
        opname_to_dataset_name = {
            'open': 'mdtargets/opens',
            'close': 'mdtargets/closes',
            'mknod': 'mdtargets/mknods',
            'link': 'mdtargets/links',
            'unlink': 'mdtargets/unlinks',
            'mkdir': 'mdtargets/mkdirs',
            'rmdir': 'mdtargets/rmdirs',
            'rename': 'mdtargets/renames',
            'getxattr': 'mdtargets/getxattrs',
            'statfs': 'mdtargets/statfss',
            'setattr': 'mdtargets/setattrs',
            'getattr': 'mdtargets/getattrs',
        }
        dataset_names = list(opname_to_dataset_name.values())

        self.init_datasets(dataset_names, lmtdb.mds_names)

        results, columns = lmtdb.get_mds_ops_data(self.query_start, self.query_end_plusplus)

        # Index the columns to speed up insertion of data
        col_map = {}
        try:
            for db_col in ['TIMESTAMP', 'MDS_ID', 'OPERATION_ID', 'SAMPLES']:
                col_map[db_col] = columns.index(db_col)
        except ValueError:
            raise ValueError("LMT database schema does not match expectation")

        # Loop through all the results of the timeseries query
        for row in results:
            if isstr(row[col_map['TIMESTAMP']]):
                # SQLite stores timestamps as a unicode string
                timestamp = datetime.datetime.strptime(row[col_map['TIMESTAMP']],
                                                       "%Y-%m-%d %H:%M:%S")
            else:
                # MySQL timestamps are automatically converted to datetime.datetime
                timestamp = row[col_map['TIMESTAMP']]

            # figure out the dataset this row's data will go into (this
            # implicitly filters out operations that aren't defined in
            # opname_to_dataset_name)
            op_name = lmtdb.mds_op_id_map[row[col_map['OPERATION_ID']]]
            dataset_name = opname_to_dataset_name.get(op_name)
            if dataset_name is None:
                continue

            # figure out which column (MDS name) this row's data will go into
            mds_name = lmtdb.mds_id_map.get(row[col_map['MDS_ID']])
            if not mds_name:
                errmsg = "unknown MDS_ID %s" % row[col_map['MDS_ID']]
                warnings.warn(errmsg)
                continue

            self[dataset_name].insert_element(
                timestamp,
                mds_name,
                row[col_map['SAMPLES']])

    def archive_oss_data(self, lmtdb):
        """Extract and encode data from LMT's OSS_DATA table

        Queries the LMT database, interprets resulting rows, and populates a
        dictionary of TimeSeries objects with those values.

        Args:
            lmtdb (LmtDb): database object
        """

        dataset_names = [
            'dataservers/cpuload',
            'dataservers/memused',
        ]

        self.init_datasets(dataset_names, lmtdb.oss_names)

        # Now query the OSS_DATA table to get byte counts over the query time range
        results, columns = lmtdb.get_oss_data(self.query_start, self.query_end_plusplus)

        # Index the columns to speed up insertion of data
        col_map = {}
        try:
            for db_col in ['TIMESTAMP', 'OSS_ID', 'PCT_CPU', 'PCT_MEMORY']:
                col_map[db_col] = columns.index(db_col)
        except ValueError:
            raise ValueError("LMT database schema does not match expectation")

        # Loop through all the results of the timeseries query
        for row in results:
            if isstr(row[col_map['TIMESTAMP']]):
                # SQLite stores timestamps as a unicode string
                timestamp = datetime.datetime.strptime(row[col_map['TIMESTAMP']],
                                                       "%Y-%m-%d %H:%M:%S")
            else:
                # MySQL timestamps are automatically converted to datetime.datetime
                timestamp = row[col_map['TIMESTAMP']]
            target_name = lmtdb.oss_id_map[row[col_map['OSS_ID']]]
            for dataset_name in dataset_names:
                target_dbcol = self.config[dataset_name].get('column')
                # target_dbcol=PCT_CPU, target_name=snx11025n022
                if target_dbcol is not None:
                    self[dataset_name].insert_element(
                        timestamp,
                        target_name,
                        row[col_map[target_dbcol]])
                else:
                    errmsg = "%s in self.config but missing 'column' setting" % dataset_name
                    raise KeyError(errmsg)

    def archive_ost_data(self, lmtdb):
        """Extract and encode data from LMT's OST_DATA table

        Queries the LMT database, interprets resulting rows, and populates a
        dictionary of TimeSeries objects with those values.

        Args:
            lmtdb (LmtDb): database object
        """

        dataset_names = [
            'datatargets/readbytes',
            'datatargets/writebytes',
            'fullness/bytes',
            'fullness/bytestotal',
            'fullness/inodes',
            'fullness/inodestotal'
        ]

        self.init_datasets(dataset_names, lmtdb.ost_names)

        # Now query the OST_DATA table to get byte counts over the query time range
        results, columns = lmtdb.get_ost_data(self.query_start, self.query_end_plusplus)

        # Index the columns to speed up insertion of data
        col_map = {}
        try:
            for db_col in ['TIMESTAMP', 'OST_ID', 'READ_BYTES',
                           'WRITE_BYTES', 'KBYTES_USED', 'KBYTES_FREE',
                           'INODES_USED', 'INODES_FREE']:
                col_map[db_col] = columns.index(db_col)
        except ValueError:
            raise ValueError("LMT database schema does not match expectation")

        # Loop through all the results of the timeseries query
        for row in results:
            if isstr(row[col_map['TIMESTAMP']]):
                # SQLite stores timestamps as a unicode string
                timestamp = datetime.datetime.strptime(row[col_map['TIMESTAMP']],
                                                       "%Y-%m-%d %H:%M:%S")
            else:
                # MySQL timestamps are automatically converted to datetime.datetime
                timestamp = row[col_map['TIMESTAMP']]
            target_name = lmtdb.ost_id_map[row[col_map['OST_ID']]]
            for dataset_name in dataset_names:
                target_dbcol = self.config[dataset_name].get('column')
                if target_dbcol is not None:
                    self[dataset_name].insert_element(
                        timestamp,
                        target_name,
                        row[col_map[target_dbcol]])
                elif dataset_name == 'fullness/bytestotal':
                    self[dataset_name].insert_element(
                        timestamp,
                        target_name,
                        row[col_map['KBYTES_USED']] + row[col_map['KBYTES_FREE']])
                elif dataset_name == 'fullness/inodestotal':
                    self[dataset_name].insert_element(
                        timestamp,
                        target_name,
                        row[col_map['INODES_USED']] + row[col_map['INODES_FREE']])
                else:
                    errmsg = "%s in self.config but missing 'column' setting" % dataset_name
                    raise KeyError(errmsg)

def init_hdf5_file(datasets, init_start, init_end, hdf5_file):
    """
    Initialize the datasets at full dimensions in the HDF5 file if necessary
    """
    schema = tokio.connectors.hdf5.SCHEMA.get(SCHEMA_VERSION)
    for dataset_name, dataset in datasets.items():
        hdf5_dataset_name = schema.get(dataset_name)
        if hdf5_dataset_name is None:
            if '/_' not in dataset_name:
                warnings.warn("Dataset key %s is not in schema" % dataset_name)
            continue
        if hdf5_dataset_name not in hdf5_file:
            # attempt to convert dataset into a timeseries
            timeseries = hdf5_file.to_timeseries(dataset_name=hdf5_dataset_name)

            # if dataset -> timeseries failed, create and commit a new, empty timeseries
            if timeseries is None:
                timeseries = tokio.timeseries.TimeSeries(dataset_name=hdf5_dataset_name,
                                                         start=init_start,
                                                         end=init_end,
                                                         timestep=dataset.timestep,
                                                         num_columns=dataset.dataset.shape[1])
                hdf5_file.commit_timeseries(timeseries=timeseries)
            print("Initialized %s in %s with size %s" % (
                hdf5_dataset_name,
                hdf5_file.name,
                timeseries.dataset.shape))

def archive_lmtdb(lmtdb, init_start, init_end, timestep, output_file, query_start, query_end):
    """
    Given a start and end time, retrieve all of the relevant contents of an LMT
    database.
    """
    datasets = DatasetDict(query_start, query_end, timestep)

    datasets.archive_ost_data(lmtdb)
    datasets.archive_oss_data(lmtdb)
    datasets.archive_mds_data(lmtdb)
    datasets.archive_mds_ops_data(lmtdb)

    datasets.finalize()

    with tokio.connectors.hdf5.Hdf5(output_file) as hdf5_file:
        hdf5_file.attrs['version'] = SCHEMA_VERSION

        init_hdf5_file(datasets, init_start, init_end, hdf5_file)

        for dataset in datasets.values():
            print("Writing out %s" % dataset.dataset_name)
            hdf5_file.commit_timeseries(dataset)

    tokio.debug.debug_print("Wrote output to %s" % output_file)

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, default=None,
                        help="input cache db file")
    parser.add_argument("-o", "--output", type=str, default='output.hdf5',
                        help="output file (default: output.hdf5)")
    parser.add_argument('--init-start', type=str, default=None,
                        help='first timestamp (inclusive) when creating new output file,' +
                        ' in %s format (default: same as start)' % DATE_FMT_PRINT)
    parser.add_argument('--init-end', type=str, default=None,
                        help='final timestamp (exclusive) when creating new output file,' +
                        ' in %s format (default: same as end)' % DATE_FMT_PRINT)
    parser.add_argument('--debug', action='store_true', help="produce debug messages")
    parser.add_argument('--timestep', type=int, default=5,
                        help='collection frequency, in seconds (default: 5)')
    parser.add_argument("--host", type=str, default=None, help="database hostname")
    parser.add_argument("--user", type=str, default=None, help="database user")
    parser.add_argument("--password", type=str, default=None, help="database password")
    parser.add_argument("--database", type=str, default=None, help="database name")
    parser.add_argument("query_start", type=str, help="start time in %s format" % DATE_FMT_PRINT)
    parser.add_argument("query_end", type=str, help="end time in %s format" % DATE_FMT_PRINT)
    args = parser.parse_args(argv)

    if args.debug:
        tokio.debug.DEBUG = True

    # Convert CLI options into datetime
    try:
        query_start = datetime.datetime.strptime(args.query_start, DATE_FMT)
        query_end = datetime.datetime.strptime(args.query_end, DATE_FMT)
        init_start = query_start
        init_end = query_end
        if args.init_start:
            init_start = datetime.datetime.strptime(args.init_start, DATE_FMT)
        if args.init_end:
            init_end = datetime.datetime.strptime(args.init_end, DATE_FMT)
    except ValueError:
        sys.stderr.write("Start and end times must be in format %s\n" % DATE_FMT)
        raise

    # Basic input bounds checking
    if query_start >= query_end:
        raise Exception('query_start >= query_end')
    elif init_start >= init_end:
        raise Exception('query_start >= query_end')
    elif args.timestep < 1:
        raise Exception('--timestep must be > 0')

    if args.input is not None:
        lmtdb = tokio.connectors.lmtdb.LmtDb(cache_file=args.input)
    else:
        lmtdb = tokio.connectors.lmtdb.LmtDb(
            dbhost=args.host,
            dbuser=args.user,
            dbpassword=args.password,
            dbname=args.database)

    archive_lmtdb(lmtdb=lmtdb,
                  init_start=init_start,
                  init_end=init_end,
                  timestep=args.timestep,
                  output_file=args.output,
                  query_start=query_start,
                  query_end=query_end)



"""
Compare two NERSC ISDCT dumps and report
    1. the devices that appeared or were removed
    2. the numeric counters whose values changed
    3. the string counters whose contents changed
"""

import json
import argparse
import warnings
import datetime
import tokio.connectors.nersc_isdct
from tokio.common import isstr

# If the following keys report changes, the drive should be flagged
ERROR_KEYS = [
    'crc_error_count',
    'critical_warnings',
    'end_to_end_error_detection_count',
    'erase_fail_count',
    'media_errors',
    'number_of_error_info_log_entries',
    'pci_link_gen_speed',
    'pci_link_width',
    'program_fail_count',
    'smart_crc_error_count_raw',
    'smart_endto_end_error_detection_count_raw',
    'smart_erase_fail_count_raw',
    'smart_pli_lock_loss_count_raw',
    'smart_program_fail_count_raw',
    'thermal_throttle_count',
]

def reduce_diff(diff_dict):
    """
    Take the raw output of .diff() and aggregate the results of each device
    """
    reduced = {
        'sum': {},
        'min': {},
        'max': {},
        'count': {},
    }
    for counters in diff_dict['devices'].values():
        for counter, value in counters.items():
            if counter not in reduced['count']:
                reduced['count'][counter] = 1
                new = True
            else:
                reduced['count'][counter] += 1
                new = False

            if not isstr(value):
                if new:
                    reduced['sum'][counter] = value
                    reduced['min'][counter] = value
                    reduced['max'][counter] = value
                else:
                    reduced['sum'][counter] += value
                    reduced['min'][counter] = min(value, reduced['min'][counter])
                    reduced['max'][counter] = max(value, reduced['max'][counter])

    result = {}
    for reduction, counters in reduced.items():
        for counter, value in counters.items():
            reduced_key = "%s_%s" % (reduction, counter)
            result[reduced_key] = value
            if reduction == 'sum':
                reduced_key = "%s_%s" % ('ave', counter)
                result[reduced_key] = float(value) / reduced['count'][counter]

    return result

def _convert_counters(counters, conversion_factor, label):
    """
    Convert a single flat dictionary of counters of bytes into another unit
    """
    results = {}

    # convert each relevant key
    for counter, value in counters.items():
        if counter.endswith('_bytes'):
            new_key = counter.replace('_bytes', "_" + label)
            new_value = value * conversion_factor
        else:
            new_key = counter
            new_value = value
        results[new_key] = new_value

    return results

def convert_byte_keys(input_dict, conversion_factor=2.0**(-30.0), label="gibs"):
    """
    Convert all keys ending in _bytes to some other unit.  Accepts either the
    raw diff dict or the reduced dict from reduce_diff()
    """
    results = {}
    # raw diff dict
    if 'devices' in input_dict:
        results = input_dict.copy()
        for serial_no, counters in input_dict['devices'].items():
            results['devices'][serial_no] = _convert_counters(counters, conversion_factor, label)
    else:
        results = _convert_counters(input_dict, conversion_factor, label)

    return results

def summarize_reduced_diffs(reduced_diffs):
    """
    Print a human-readable summary of the relevant reduced diff data
    """
    buf = ""
    ### General summary
    if 'sum_data_units_read_gibs' not in reduced_diffs:
        read_gibs = reduced_diffs.get('sum_data_units_read_bytes', 0) * 2.0**(-40)
        write_gibs = reduced_diffs.get('sum_data_units_written_bytes', 0) * 2.0**(-40)
    else:
        read_gibs = reduced_diffs.get('sum_data_units_read_gibs', 0)
        write_gibs = reduced_diffs.get('sum_data_units_written_gibs', 0)

    buf += "Read:    %10.2f TiB, %10.2f MOps\n" % (
        read_gibs,
        reduced_diffs.get('sum_host_read_commands', 0) / 1000000.0)
    buf += "Written: %10.2f TiB, %10.2f MOps\n" % (
        write_gibs,
        reduced_diffs.get('sum_host_write_commands', 0) / 1000000.0)
    buf += "WAF:     %+10.4f\n" % reduced_diffs.get('max_write_amplification_factor', 0)
    return buf

def summarize_errors(diff_dict, isdct_data):
    """
    Print a human-readable summary of any bad SSDs
    """
    buf = ""
    for serial_no, error_key in discover_errors(diff_dict):
        if buf != "":
            buf += "\n"
        buf += "%s %s %s %s" % (
            isdct_data[serial_no]['node_name'],
            serial_no,
            error_key,
            diff_dict['devices'][serial_no][error_key])
    return buf

def discover_errors(diff_dict):
    """
    Look through all diffs and report serial numbers of devices that show
    changes in counters that may indicate a hardware issue.
    """
    if 'devices' not in diff_dict:
        warnings.warn("No 'devices' found in diff dict")
        return []

    errors = []
    for serial_no, counters in diff_dict['devices'].items():
        for error_key in ERROR_KEYS:
            if error_key in counters:
                errors.append((serial_no, error_key))

    return errors

def print_summary(old_isdctfile, new_isdctfile, diff_dict):
    """Print a human-readable summary of diff_dict
    """
    reduced_diff = reduce_diff(diff_dict)
    diff_buf = summarize_reduced_diffs(reduced_diff)
    err_buf = summarize_errors(diff_dict, new_isdctfile)

    print("=== ISDCT Summary: %s - %s ===" % (
        datetime.datetime.fromtimestamp(next(iter(old_isdctfile.values()))['timestamp']),
        datetime.datetime.fromtimestamp(next(iter(new_isdctfile.values()))['timestamp'])))

    if len(err_buf) > 0:
        print("\n*** Errors Detected! ***")
        print(err_buf)

    if len(diff_dict['removed_devices']) > 0:
        print("\n=== Devices Removed ===")
        for dev in diff_dict['removed_devices']:
            print("%s %s" % (old_isdctfile[dev]['node_name'], dev))

    if len(diff_dict['added_devices']) > 0:
        print("\n=== Devices Installed ===")
        for dev in diff_dict['added_devices']:
            print("%s %s" % (new_isdctfile[dev]['node_name'], dev))

    if len(diff_buf) > 0:
        print("\n=== Workload Statistics ===")
        print(diff_buf)

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--all", action='store_true',
                        help='report changes for each device')
    parser.add_argument("-g", "--gibs", action="store_true",
                        help="report in units of GiB")
    parser.add_argument("-s", "--summary", action="store_true",
                        help="print summary of differences")
    parser.add_argument("-z", "--report-zeros", action='store_true',
                        help='include counters that do not change')
    parser.add_argument("old_isdctfile", help="older ISDCT dump file")
    parser.add_argument("new_isdctfile", help="newer ISDCT dump file")
    args = parser.parse_args(argv)

    old_isdctfile = tokio.connectors.nersc_isdct.NerscIsdct(args.old_isdctfile)
    new_isdctfile = tokio.connectors.nersc_isdct.NerscIsdct(args.new_isdctfile)
    diff_dict = new_isdctfile.diff(old_isdctfile, report_zeros=args.report_zeros)

    if args.gibs:
        diff_dict = convert_byte_keys(diff_dict)

    if args.summary:
        print_summary(old_isdctfile, new_isdctfile, diff_dict)
    elif args.all:
        print(json.dumps(diff_dict, indent=4, sort_keys=True))
    else:
        reduced_diff = reduce_diff(diff_dict)
        print(json.dumps(reduced_diff, indent=4, sort_keys=True))



"""
pytokio implements its command-line tools within this package.  Each such CLI
tool either implements some useful analysis on top of pytokio connectors,
tools, or analysis or maps some of the internal python APIs to command-line
arguments.

Most of these tools implement a ``--help`` option to explain the command-line
options.
"""



"""
Provides CLI interfaces for :meth:`tokio.connectors.nersc_jobsdb.NerscJobsDb.get_concurrent_jobs`
"""

import os
import time
import datetime
import argparse
import tokio.connectors.nersc_jobsdb

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("start", type=str, help="start time in YYYY-MM-DDTHH:MM:SS format")
    parser.add_argument("end", type=str, help="end time in YYYY-MM-DDTHH:MM:SS format")
    parser.add_argument("host", type=str, help="return jobs running on this NERSC host")
    parser.add_argument("-i", "--input", type=str, default=None, help="input cache db file")
    parser.add_argument("-o", "--output", type=str, default=None, help="output file")
    args = parser.parse_args(argv)

    start = datetime.datetime.strptime(args.start, "%Y-%m-%dT%H:%M:%S")
    end = datetime.datetime.strptime(args.end, "%Y-%m-%dT%H:%M:%S")

    if args.input is not None:
        nerscjobsdb = tokio.connectors.nersc_jobsdb.NerscJobsDb(cache_file=args.input)
    else:
        nerscjobsdb = tokio.connectors.nersc_jobsdb.NerscJobsDb()
    nerscjobsdb.get_concurrent_jobs(
        int(time.mktime(start.timetuple())),
        int(time.mktime(end.timetuple())),
        args.host)

    cache_file = args.output
    if cache_file is None:
        i = 0
        while True:
            cache_file = "nersc_jobsdb-%d.sqlite" % i
            if os.path.exists(cache_file):
                i += 1
            else:
                break
    print("Caching to %s" % cache_file)
    nerscjobsdb.save_cache(cache_file)



#/usr/bin/env python
"""
TimeSeries class to simplify updating and manipulating the in-memory
representation of time series data.
"""

import re
import math
import time
import datetime
import warnings
import numpy
from tokio.common import isstr

class TimeSeries(object):
    """
    In-memory representation of an HDF5 group in a TokioFile.  Can either
    initialize with no datasets, or initialize against an existing HDF5
    group.
    """
    def __init__(self, dataset_name=None,
                 start=None, end=None, timestep=None, num_columns=None,
                 column_names=None, timestamp_key=None,
                 sort_hex=False):

        # numpy.ndarray of timestamp measurements
        self.timestamps = None

        # time between consecutive timestamps
        self.timestep = None
        # numpy.ndarray of the timeseries data itself
        self.dataset = None
        # string containing fully qualified dataset name+path
        self.dataset_name = None
        # list of strings serving as column headings
        self.columns = []
        # map between strings in self.columns and its corresponding column index in self.dataset
        self.column_map = {}
        # key-value of metadata to be stored in HDF5 group.attrs
        self.group_metadata = {}
        # key-value of metadata to be stored in HDF5 dataset.attrs
        self.dataset_metadata = {}
        # path corresponding to self.timestamps dataset in HDF5 file
        self.timestamp_key = timestamp_key
        # True = natural sort columns assuming hex-encoded numbers; False = only recognize decimals
        self.sort_hex = sort_hex
        # string describing dataset version
        self.version = None
        # string describing schema version
        self.global_version = None

        # attempt to initialize the object if fields are supplied
        if dataset_name is not None and start and end and timestep and num_columns:
            self.init(start, end, timestep, num_columns, dataset_name,
                column_names, timestamp_key)

    def init(self, start, end, timestep, num_columns, dataset_name,
             column_names=None, timestamp_key=None):
        """Create a new TimeSeries dataset object

        Responsible for setting self.timestep, self.timestamp_key, and self.timestamps

        Args:
            start (datetime): timestamp to correspond with the 0th index
            end (datetime): timestamp at which timeseries will end (exclusive)
            timestep (int): seconds between consecutive timestamp indices
            num_columns (int): number of columns to initialize in the numpy.ndarray
            dataset_name (str): an HDF5-compatible name for this timeseries
            column_names (list of str, optional): strings by which each column
                should be indexed.  Must be less than or equal to num_columns in
                length; difference remains uninitialized
            timestamp_key (str, optional): an HDF5-compatible name for this timeseries'
                timestamp vector.  Default is /groupname/timestamps
        """
        if column_names is None:
            column_names = []

        # Attach the timestep dataset
        self.timestep = timestep

        # Calculate the hours in a day in epoch-seconds since Python datetime
        # and timedelta doesn't understand DST
        time_list = []
        end_epoch = int(time.mktime(end.timetuple()))
        timestamp = int(time.mktime(start.timetuple()))
        while timestamp < end_epoch:
            time_list.append(timestamp)
            timestamp += timestep

        self.timestamps = numpy.array(time_list)

        # Attach the dataset itself
        self.dataset_name = dataset_name
        self.dataset = numpy.full((len(self.timestamps), num_columns), -0.0)
        self.set_columns(column_names)

        self.set_timestamp_key(timestamp_key, safe=True)

    def set_timestamp_key(self, timestamp_key, safe=False):
        """Set the timestamp key

        Args:
            timestamp_key (str): The key for the timestamp dataset.
            safe (bool): If true, do not overwrite an existing timestamp key
        """
        # Root the timestamp_key at the same parent as the dataset
        if not safe or self.timestamp_key is None:
            self.timestamp_key = timestamp_key

    def update_column_map(self):
        """
        Create the mapping of column names to column indices
        """
        self.column_map = {}
        for index, column_name in enumerate(self.columns):
            self.column_map[column_name] = index

    def set_columns(self, column_names):
        """
        Set the list of column names
        """
        num_columns = len(column_names)
        # handle case where HDF5 has more column names than dataset columns via truncation
        if num_columns > self.dataset.shape[1]:
            column_names = self.columns[0:self.dataset.shape[1]]
            truncated = self.columns[self.dataset.shape[1]:]
            warnings.warn(
                "Dataset has %d column names but %d columns; dropping columns %s"
                % (num_columns, self.dataset.shape[1], ', '.join(truncated)))
        self.columns = [str(x) for x in column_names] # ensure that elements are mutable and not unicode
        self.update_column_map()

    def add_column(self, column_name):
        """
        Add a new column and update the column map
        """
        index = len(self.columns)
        if column_name in self.column_map:
            warnings.warn("Adding degenerate column '%s' at %d (exists at %d)"
                          % (column_name, index, self.column_map[column_name]))
        self.column_map[column_name] = index
        if index >= (self.dataset.shape[1]):
            errmsg = "new index %d (%s) exceeds number of columns %d (%s)" % (
                index, column_name, self.dataset.shape[1], self.columns[-1])
            raise IndexError(errmsg)
        self.columns.append(str(column_name)) # convert from unicode to str for numpy
        return index

    def sort_columns(self):
        """
        Rearrange the dataset's column data by sorting them by their headings
        """
        self.rearrange_columns(sorted_nodenames(self.columns, sort_hex=self.sort_hex))

    def rearrange_columns(self, new_order):
        """
        Rearrange the dataset's columnar data by an arbitrary column order given
        as an enumerable list
        """
        # validate the new order - new_order must contain at least all of
        # the elements in self.columns, but may contain more than that
        for new_key in new_order:
            if new_key not in self.columns:
                raise Exception("key %s in new_order not in columns" % new_key)

        # walk the new column order
        for new_index, new_column in enumerate(new_order):
            # new_order can contain elements that don't exist; this happens when
            # re-ordering a small dataset to be inserted into an existing,
            # larger dataset
            if new_column not in self.columns:
                warnings.warn("Column '%s' in new order not present in TimeSeries" % new_column)
                continue

            old_index = self.column_map[new_column]
            self.swap_columns(old_index, new_index)

    def swap_columns(self, index1, index2):
        """
        Swap two columns of the dataset in-place
        """
        # save the data from the column we're about to swap
        saved_column_data = self.dataset[:, index2].copy()
        saved_column_name = self.columns[index2]

        # swap column data
        self.dataset[:, index2] = self.dataset[:, index1]
        self.dataset[:, index1] = saved_column_data[:]

        # swap column names too
        self.columns[index2] = self.columns[index1]
        self.columns[index1] = saved_column_name

        # update the column map
        self.column_map[self.columns[index2]] = index2
        self.column_map[self.columns[index1]] = index1

    def get_insert_pos(self, timestamp, column_name, create_col=False):
        """Determine col and row indices corresponding to timestamp and col name

        Args:
            timestamp (datetime): Timestamp to map to a row index
            column_name (str): Name of column to map to a column index
            create_col (bool): If column_name does not exist, create it?
        Returns:
            (t_index, c_index) (long or None)
        """
        timestamp_epoch = int(time.mktime(timestamp.timetuple()))
        t_index = (timestamp_epoch - self.timestamps[0]) // self.timestep
        if t_index >= self.timestamps.shape[0]: # check bounds
            return None, None

        # create a new column label if necessary
        c_index = self.column_map.get(column_name)
        if c_index is None and create_col:
            c_index = self.add_column(column_name)
        return t_index, c_index

    def insert_element(self, timestamp, column_name, value, reducer=None):
        """
        Given a timestamp (datetime.datetime object) and a column name (string),
        update an element of the dataset.  If a reducer function is provided,
        use that function to reconcile any existing values in the element to be
        updated.
        """
        t_index, c_index = self.get_insert_pos(timestamp,
                                               column_name,
                                               create_col=True)
        if t_index is None or c_index is None:
            return False

        # actually copy the two data points into the datasets
        old_value = self.dataset[t_index, c_index]
        if (self.dataset[t_index, c_index] != 0.0 \
        or math.copysign(1, old_value) < 0.0) \
        and reducer is not None:
            self.dataset[t_index, c_index] = reducer(old_value, value)
        else:
            self.dataset[t_index, c_index] = value
        return True

    def convert_to_deltas(self):
        """
        Convert a matrix of monotonically increasing rows into deltas.  Replaces
        self.dataset with a matrix with the same number of columns but one fewer
        row (taken off the bottom of the matrix).  Also adjusts the timestamps
        dataset.
        """
        self.dataset = timeseries_deltas(self.dataset)
        self.timestamps = self.timestamps[0:-1]

    def trim_rows(self, num_rows=1):
        """
        Trim some rows off the end of self.dataset and self.timestamps
        """
        self.dataset = self.dataset[0:-1*num_rows]
        self.timestamps = self.timestamps[0:-1*num_rows]

    def add_rows(self, num_rows=1):
        """
        Add additional rows to the end of self.dataset and self.timestamps
        """
        new_dataset_rows = numpy.full((num_rows, self.dataset.shape[1]), -0.0)
        new_timestamps = []
        for index in range(num_rows):
            new_timestamps.append(self.timestamps[-1] + (index + 1) * self.timestep)
        new_timestamp_rows = numpy.array(new_timestamps)

        self.dataset = numpy.vstack((self.dataset, new_dataset_rows))
        self.timestamps = numpy.hstack((self.timestamps, new_timestamp_rows))

def sorted_nodenames(nodenames, sort_hex=False):
    """
    Gnarly routine to sort nodenames naturally.  Required for nodes named things
    like 'bb23' and 'bb231'.
    """
    def extract_int(string):
        """
        Convert input into an int if possible; otherwise return unmodified
        """
        try:
            if sort_hex:
                return int(string, 16)
            return int(string)
        except ValueError:
            return string

    def natural_compare(string):
        """
        Tokenize string into alternating strings/ints if possible
        """
        return list(map(extract_int, re.findall(r'(\d+|\D+)', string)))

    def natural_hex_compare(string):
        """
        Tokenize string into alternating strings/ints if possible.  Also
        recognizes hex, so be careful with ambiguous nodenames like "bb234",
        which is valid hex.
        """
        return list(map(extract_int, re.findall(r'([0-9a-fA-F]+|[^0-9a-fA-F]+)', string)))

#   def natural_comp(arg1, arg2):
#       """
#       Cast the parts of a string that look like integers into integers, then
#       sort based on strings and integers rather than only strings
#       """
#       return cmp(natural_compare(arg1), natural_compare(arg2))

#   def natural_hex_comp(arg1, arg2):
#       """
#       Cast the parts of a string that look like hex into integers, then
#       sort based on strings and integers rather than only strings.
#       """
#       return cmp(natural_hex_compare(arg1), natural_hex_compare(arg2))

    if sort_hex:
        return sorted(nodenames, key=natural_hex_compare)
    return sorted(nodenames, key=natural_compare)

def timeseries_deltas(dataset):
    """Convert monotonically increasing values into deltas

    Subtract every row of the dataset from the row that precedes it to
    convert a matrix of monotonically increasing rows into deltas.  This is a
    lossy process because the deltas for the final measurement of the time
    series cannot be calculated.

    Args:
        dataset (numpy.ndarray): The dataset to convert from absolute values
            into deltas.  rows should correspond to time, and columns to
            individual components

    Returns:
        numpy.ndarray: The deltas between each row in the given input dataset.
            Will have the same number of columns as the input dataset and one
            fewer rows.
    """
    diff_matrix = numpy.full((dataset.shape[0] - 1, dataset.shape[1]), -0.0)

    prev_nonzero = [None] * dataset.shape[1] # the last known valid measurement
    searching = [True] * dataset.shape[1] # are we spanning a gap in data?
    for irow in range(dataset.shape[0]):
        for icol in range(dataset.shape[1]):
            this_element = dataset[irow, icol]

            if irow == 0:
                if this_element != 0.0:
                    prev_nonzero[icol] = this_element
            elif searching[icol]:
                if this_element != 0.0:
                    if prev_nonzero[icol] is not None and this_element >= prev_nonzero[icol]:
                        diff_matrix[irow - 1, icol] = this_element - prev_nonzero[icol]
                        searching[icol] = False
                    prev_nonzero[icol] = this_element
            else:
                if this_element < dataset[irow - 1, icol]: # found a missing data point
                    searching[icol] = True
                else:
                    diff_matrix[irow - 1, icol] = this_element - dataset[irow - 1, icol]
                    prev_nonzero[icol] = this_element

    return diff_matrix



"""pytokio is a framework that enables holistic analysis of telemetry from HPC systems

pytokio is a reference implementation of the Total Knowledge of I/O (TOKIO)
framework designed to simplify the process of holistically analyzing the
performance of parallel storage systems commonly deployed in high-performance
computing.  It provides connectors to interface with many common monitoring
tools, analysis routines to derive insight from multiple tools, and example
applications that demonstrate simple but powerful performance analyses enabled
by this holistic approach.

pytokio is made available under a modified BSD license.  See
https://pytokio.readthedocs.io/en/latest/ for the full documentation.
"""

#
#  To install: python setup.py install
#              python setup.py install --prefix=/path/to/prefix
#
#  To package: python setup.py sdist
#                              bdist
#                              bdist_egg
#
#  Making bdists is not supported because site.json cannot be correctly
#  resolved.  If you choose to use a bdist, you must explicitly set
#  PYTOKIO_CONFIG to point to a valid site.json.
#
import os
import re
import glob
import subprocess

RELEASE = True # set to True when building a release distribution


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
DOCLINES = (__doc__ or '').split("\n")
REQUIREMENTS = open(os.path.join(BASE_DIR, 'requirements.txt'), 'r').readlines()

if os.path.exists('MANIFEST'):
    os.remove('MANIFEST')

def setup_package():
    import setuptools

    METADATA = dict(
        name='pytokio',
        version=find_version(),
        author='Glenn K. Lockwood et al.',
        author_email='glock@lbl.gov',
        description=DOCLINES[0],
        long_description="\n".join(DOCLINES[2:]),
        url="http://www.nersc.gov/research-and-development/tokio/",
        download_url="https://www.github.com/nersc/pytokio",
        license='BSD',
        platforms=["Linux", "MacOS-X"],

        # Dependencies
        install_requires=REQUIREMENTS,
        extras_require={
            'collectdes': ['elasticsearch>=5.4'],
            'nersc_globuslogs': ['elasticsearch>=5.4'],
            'esnet_snmp': ['requests'],
            'lmtdb': ['mysqlclient'],
            'nersc_jobsdb': ['mysqlclient'],
        },
        python_requires=">=2.7",

        classifiers=[
            'Intended Audience :: Science/Research',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3',
            'Topic :: Software Development',
            'Topic :: Scientific/Engineering',
            'Operating System :: POSIX',
            'Operating System :: Unix',
            'Operating System :: MacOS',
        ],
        keywords='I/O performance monitoring'
    )

    # Include the tokio package
    METADATA['packages'] = setuptools.find_packages()

    # Create the CLI interface
    METADATA['entry_points'] = {'console_scripts': generate_console_scripts()}

    # If we want to keep site.json at the top-level and copy it in during
    # install time.  This would force users to correctly install pytokio
    # before it could be used though, which is not strictly necessary for
    # any other purpose.
    # METADATA['data_files'] = [('etc', ['site.json'])]

    # Include all non-py files in MANIFEST['packages'] (including site.json)
    METADATA['include_package_data'] = True

    setuptools.setup(**METADATA)

def find_version():
    """Extract the package version number from __init__.py

    See https://packaging.python.org/guides/single-sourcing-package-version
    """
    init_file = os.path.join(BASE_DIR, "tokio", "__init__.py")
    if os.path.isfile(init_file):
        match = re.search(r"^__version__\s*=\s*['\"]([^'\"]+?)['\"]",
                          open(init_file, 'r').read(),
                          re.M)
        if match:
            version = match.group(1)
            if not RELEASE:
                revision_cache = os.path.join(BASE_DIR, '.revision')
                revision = git_version()
                if revision == "unknown" and os.path.isfile(revision_cache):
                    with open(revision_cache, 'r') as revision_cache_f:
                        revision = revision_cache_f.read().strip()
                if revision and revision != "unknown":
                    version = "%s+%s" % (version, revision[:7])
                    with open(revision_cache, 'w') as revision_cache_f:
                        revision_cache_f.write(revision)
                else:
                    version = "%s+unknown" % (version)
            return version
        else:
            raise RuntimeError("Unable to find version string")
    else:
        raise RuntimeError("Unable to find version string")


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH', 'HOME']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        revision = out.strip().decode('ascii')
    except OSError:
        revision = "unknown"

    return revision

def generate_console_scripts():
    """Generates the console_scripts argument to setuptools.setup

    Generates the console_scripts key:value to be passed into setuptools.setup(entry_points=...)
    """
    console_scripts = []
    for x in glob.glob(os.path.join(BASE_DIR, 'tokio', 'cli', '*.py')):
        if '__init__' not in x:
            basename = os.path.basename(x).rsplit('.', 1)[0]
            console_scripts.append("%s = tokio.cli.%s:main" % (basename, basename))

    print("Including scripts:")
    print('\n  '.join(console_scripts))

    return console_scripts

if __name__ == "__main__":
    setup_package()



#!/usr/bin/env python
"""
The Total Knowledge of I/O (TOKIO) reference implementation, pytokio.
"""

import tokio.common
import tokio.timeseries
import tokio.config
import tokio.debug

import tokio.connectors
import tokio.tools
import tokio.analysis
# import tokio.cli # do not import by default

# Exceptions
from tokio.common import ConfigError

__version__ = '0.12.0b5'



"""
Provides CLI interface for the :meth:`tokio.tools.darshan.load_darshanlogs` tool
which locates darshan logs in the system-wide repository.
"""

import sys
import json
import argparse
import datetime
import tokio.tools.darshan

DATE_FMT = "%Y-%m-%d"
DATE_FMT_PRINT = "YYYY-MM-DD"

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--start', type=str, default=None,
                        help='date to begin search for logs, in %s format ' % DATE_FMT_PRINT
                        + '(default: same as start)')
    parser.add_argument('-e', '--end', type=str, default=None,
                        help='date to end search for logs (inclusive),'
                        + ' in %s format (default: same as start)' % DATE_FMT_PRINT)
    parser.add_argument("-u", "--username", type=str, default=None,
                        help="username of Darshan log owner")
    parser.add_argument("-j", "--jobid", type=str, default=None,
                        help="jobid of Darshan log")
    parser.add_argument("-l", "--load", type=str, default=None,
                         help='load each Darshan log; must be {base[,total][,perf]}')
    parser.add_argument('--logdir', type=str,
                        help='path to DARSHAN_LOG_DIR, exclusive of dated subdirectories (default: use site config value)')
    parser.add_argument('--host', type=str, default=None,
                        help="hostname; only required if --logdir is not specified and site config contains multiple darshan_log_dirs")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Enable debug messages")
    args = parser.parse_args(argv)

    start = None
    end = None
    try:
        if args.start:
            start = datetime.datetime.strptime(args.start, DATE_FMT)
    
        if args.end:
            end = datetime.datetime.strptime(args.end, DATE_FMT)

    except ValueError:
        sys.stderr.write("Start and end times must be in format %s\n" % DATE_FMT_PRINT)
        raise

    # Basic input bounds checking
    if end and not start:
        start = end
    if start and not end:
        end = start
    if start and start > end:
        raise ValueError('query_start >= query_end')
    elif not start and not args.jobid:
        raise ValueError("Either --start/--end or --jobid must be defined")

    if args.verbose:
        tokio.debug.DEBUG = True

    if args.load:
        results = tokio.tools.darshan.load_darshanlogs(datetime_start=start,
                                                       datetime_end=end,
                                                       username=args.username,
                                                       jobid=args.jobid,
                                                       which=args.load,
                                                       log_dir=args.logdir,
                                                       system=args.host)
        print(json.dumps(results, indent=4, sort_keys=True))

    else:
        for logfile in tokio.tools.darshan.find_darshanlogs(datetime_start=start,
                                                            datetime_end=end,
                                                            username=args.username,
                                                            jobid=args.jobid,
                                                            log_dir=args.logdir,
                                                            system=args.host):
            print(logfile)



"""Common convenience routines used throughout pytokio
"""

import time
import json
import datetime
import numpy

UNITS = [
    (2**80, "YiB"),
    (2**70, "ZiB"),
    (2**60, "EiB"),
    (2**50, "PiB"),
    (2**40, "TiB"),
    (2**30, "GiB"),
    (2**20, "MiB"),
    (2**10, "KiB")
]
UNITS_10 = [
    (10**24, "YiB"),
    (10**21, "ZiB"),
    (10**18, "EiB"),
    (10**15, "PiB"),
    (10**12, "TiB"),
    (10**9, "GiB"),
    (10**6, "MiB"),
    (10**3, "KiB")
]

class ConfigError(RuntimeError):
    pass

def isstr(obj):
    """Determine if an object is a string or string-derivative

    Provided for Python2/3 compatibility

    Args:
        obj: object to be tested for stringiness

    Returns:
        bool: is it string-like?
    """
    try:
        # basestring is only in Python 2
        return isinstance(obj, basestring)
    except NameError:
        # Python 3
        return isinstance(obj, str)

class JSONEncoder(json.JSONEncoder):
    """Convert common pytokio data types into serializable formats
    """
    def default(self, obj): # pylint: disable=E0202
        if isinstance(obj, datetime.datetime):
            return int(time.mktime(obj.timetuple()))
        elif isinstance(obj, numpy.int64):
            return int(obj)
        return json.JSONEncoder.default(self, obj)

def humanize_bytes(bytect, base10=False, fmt="%.1f %s"):
    """Converts bytes into human-readable units

    Args:
        bytect (int): Number of bytes
        base10 (bool): Convert to base-10 units (MB, GB, etc) if True
        fmt (str or None): Format of string to return; must contain %f/%d and %s
            for the quantity and units, respectively.

    Returns:
        str: Quantity and units expressed in a human-readable quantity
    """

    if base10:
        units = UNITS_10
    else:
        units = UNITS

    for unit in units:
        if bytect >= unit[0]:
            return fmt % (bytect / unit[0], unit[1])

    return fmt % (bytect, "bytes" if bytect != 1 else "byte")

def to_epoch(datetime_obj):
    """Convert datetime.datetime into epoch seconds

    Currently assumes input datetime is expressed in localtime.  Does not handle
    timezones very well.

    Args:
        datetime_obj (datetime.datetime): Datetime to convert to seconds-since-epoch

    Returns:
        int: Seconds since epoch
    """

    return int(time.mktime(datetime_obj.timetuple()))

def recast_string(value):
    """Converts a string to some type of number or True/False if possible

    Args:
        value (str): A string that may represent an int or float

    Returns:
        int, float, bool, or str: The most precise numerical or boolean
        representation of ``value`` if ``value`` is a valid string-encoded
        version of that type.  Returns the unchanged string otherwise.
    """

    ### the order here is important, but hex that is not prefixed with
    ### 0x may be misinterpreted as integers.
    new_value = value
    for cast in (int, float, lambda x: int(x, 16)):
        try:
            new_value = cast(value)
            break
        except ValueError:
            pass

    if value == "True":
        new_value = True
    elif value == "False":
        new_value = False

    return new_value



"""
Provides CLI interface for :meth:`tokio.tools.topology.get_job_diameter`.
"""

import json
import argparse
import tokio.tools.topology

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--jobinfo-cache", default=None, type=str,
                       help="path to jobinfo (e.g., Slurm) cache file")
    parser.add_argument("--nodemap-cache", default=None, type=str,
                        help="path to xtdb2proc cache file")
    parser.add_argument("-o", "--output", type=str, default=None, help="output file")
    group.add_argument("jobid", nargs='?', default=None, help="Slurm job id of interest")
    args = parser.parse_args(argv)

    topology_result = tokio.tools.topology.get_job_diameter(jobid=args.jobid,
                                                            nodemap_cache_file=args.nodemap_cache,
                                                            jobinfo_cache_file=args.jobinfo_cache)
    # Serialize the object
    cache_file = args.output
    if cache_file is None:
        print(json.dumps(topology_result, indent=4, sort_keys=True))
    else:
        print("Caching to %s" % cache_file)
        json.dump(topology_result, open(cache_file, 'w'))



#!/usr/bin/env python
"""
Tools to parse and index the outputs of Lustre's ``lfs`` and ``lctl`` commands
to quantify Lustre fullness and health.  Assumes inputs are generated by NERSC's
Lustre health monitoring cron jobs which periodically issue the following::

    echo "BEGIN $(date +%s)" >> osts.txt
    /usr/bin/lfs df >> osts.txt

    echo "BEGIN $(date +%s)" >> ost-map.txt
    /usr/sbin/lctl dl -t >> ost-map.txt

Accepts ASCII text files, or gzip-compressed text files.
"""

import sys
import gzip
import re
import warnings
import mimetypes
import tokio.connectors.lfshealth
from tokio.common import isstr

_REX_OST_MAP = re.compile(r'^\s*(\d+)\s+(\S+)\s+(\S+)\s+(snx\d+-\S+)\s+(\S+)\s+(\d+)\s+(\S+@\S+)\s*$')
"""Regular expression to match OSC/MDC lines

Matches output of ``lctl dl -t`` which takes the form::

    351 UP osc snx11025-OST0007-osc-ffff8875ac1e7c00 3f30f170-90e6-b332-b141-a6d4a94a1829 5 10.100.100.12@o2ib1

Intentionally skips MGC, LOV, and LMV lines.
"""

_REX_LFS_DF = re.compile(r'^\s*(snx\d+-\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+).\s+(\S+)\[([^:]+):(\d+)\]\s*$')
"""Regular expression to extract OST fullness levels

Matches output of ``lfs df`` which takes the form::

    snx11035-OST0000_UUID 90767651352 54512631228 35277748388  61% /scratch2[OST:0]

where the columns are

    * OST/MDT UID
    * kibibytes total
    * kibibytes in use
    * kibibytes available
    * percent fullness
    * file system mount, role, and ID

Carries the implicit assumption that all OSTs are prefixed with `snx`.
"""

class NerscLfsOstMap(dict):
    """Subclass of dictionary that self-populates with Lustre OST-OSS mapping.
    """
    def __init__(self, cache_file=None):
        """Load the mapping of OSTs to OSSes.

        Args:
            cache_file (str, optional): Path to a cache file to load instead of
                issuing the ``lctl dl -t`` command
        """
        super(NerscLfsOstMap, self).__init__(self)
        self.cache_file = cache_file
        self.load_ost_map_file()

    def __repr__(self):
        """Serialize OST map into a format that resembles ``lctl dl -t``.

        Returns:
            str: Serialization of the OST to OSS mapping in a format similar to
            ``lctl dl -t``.  Fixed-width columns are

            * index: OST/MDT index
            * status: up/down status
            * role: ``osc``, ``mdc``, etc
            * role_id: name with unique identifier for target
            * uuid: UUID of target
            * ref_count: number of references to target
            * nid: LNET identifier of the target
        """
        repr_result = ""
        # Iterate over time steps ("BEGIN" lines)
        for timestamp in sorted(self.keys()):
            fs_data = self[timestamp]
            repr_result += "BEGIN %d\n" % timestamp
            repr_result += str(fs_data)
        return repr_result

    def load_ost_map_file(self):
        """Parse the cached output of an OST map generated by ``lctl dl -t``.

        Reads the input OST map as given by the ``cache_file`` attribute and
        populates self with keys of the form::

            { timestamp(int) : { file_system: { ost_name : { keys: values } } } }
        """
        _, encoding = mimetypes.guess_type(self.cache_file)
        if encoding == 'gzip':
            input_file = gzip.open(self.cache_file, 'rt')
        else:
            input_file = open(self.cache_file, 'r')

        this_timestamp = None
        degenerate_keys = 0
        load_str = []
        for line in input_file:
            if line.startswith('BEGIN'):
                if degenerate_keys > 0:
                    warnings.warn("%d degenerate keys found for timestamp %d" % (degenerate_keys, this_timestamp))
                degenerate_keys = 0

                # append previous data
                if load_str:
                    self.__setitem__(this_timestamp,
                                     tokio.connectors.lfshealth.LfsOstMap(from_string='\n'.join(load_str)))

                # initialize new timestamp
                this_timestamp = int(line.split()[1])
                if this_timestamp in self:
                    warnings.warn("degenerate timestamp %d found" % this_timestamp)

                load_str = []
            else:
                load_str.append(line)

        input_file.close()

    def save_cache(self, output_file=None):
        """Serialize object into a form resembling the output of ``lctl dl -t``.

        Args:
            output_file (str): Path to a file to which the serialized output
                should be written.  If None, print to stdout.
        """
        if output_file is None:
            self._save_cache(sys.stdout)
        else:
            with open(output_file, 'w') as output:
                self._save_cache(output)

    def _save_cache(self, output):
        """Serialize object into a form resembling the output of ``lctl dl -t``.

        Args:
            output (file): File-like object into which resulting text should be
                written.
        """
        output.write(str(self))

    def get_failovers(self):
        """Determine OSSes which are likely affected by a failover.

        Figure out the OSTs that are probably failed over and, for each time
        stamp and file system, return a list of abnormal OSSes and the expected
        number of OSTs per OSS.

        Returns:
            dict: Dictionary keyed by timestamps and whose values are dicts of
            the form::

                {
                    'mode': int,
                    'abnormal_ips': [list of str]
                }

            where ``mode`` refers to the statistical mode of OSTs per OSS, and
            ``abnormal_ips`` is a list of strings containing the IP addresses of
            OSSes whose OST counts are not equal to the ``mode`` for that
            time stamp.
        """
        resulting_data = {}
        for timestamp, fs_data in self.items():
            resulting_data[timestamp] = fs_data.get_failovers()

        return resulting_data

class NerscLfsOstFullness(dict):
    """Subclass of dictionary that self-populates with Lustre OST fullness.
    """
    def __init__(self, cache_file=None):
        """Load the fullness of OSTs

        Args:
            cache_file (str, optional): Path to a cache file to load instead of
                issuing the ``lfs df`` command
        """
        super(NerscLfsOstFullness, self).__init__(self)
        self.cache_file = cache_file
        self.load_ost_fullness_file()

    def __repr__(self):
        """Serialize OST fullness into a format that resembles ``lfs df``.

        Returns:
            str: Serialization of the OST fullness in a format similar to
            ``lfs df``.  Columns are

            * Name of OST (e.g., snx11025-OST0001_UUID)
            * Total kibibytes on OST
            * Used kibibytes on OST
            * Available kibibytes on OST
            * Percent capacity used
            * Mount point, role, and OST ID
        """
        repr_result = ""
        for timestamp in sorted(self.keys()):
            repr_result += "BEGIN %d\n" % timestamp
            fs_data = self[timestamp]
            for target_name in sorted(fs_data.keys()):
                obd_data = fs_data[target_name]
                for obd_name in sorted(list(obd_data.keys()),
                                       key=lambda x, y=obd_data: y[x]['target_index']):
                    keyvalues = obd_data[obd_name]
                    repr_result += "%s-%s_UUID %ld %ld %ld %3d%% %s[%s:%d]\n" % (
                        target_name,
                        obd_name,
                        keyvalues['total_kib'],
                        keyvalues['used_kib'],
                        keyvalues['remaining_kib'],
                        # Note that lfs dl's percents are not divided by
                        # avail_kib, but rather the sum of used and remaining.
                        round(100.0 * keyvalues['used_kib'] / (keyvalues['remaining_kib']
                                                               + keyvalues['used_kib'])),
                        keyvalues['mount_pt'],
                        keyvalues['role'].upper(),
                        keyvalues['target_index'], )

        return repr_result

    def load_ost_fullness_file(self):
        """Parse the cached output of OST fullness generated by ``lfs df``.

        Parses the output of a file containing concatenated outputs of `lfs df`
        separated by lines of the form `BEGIN 0000` where 0000 is the UNIX epoch
        time.
        """
        _, encoding = mimetypes.guess_type(self.cache_file)
        if encoding == 'gzip':
            input_file = gzip.open(self.cache_file, 'rt')
        else:
            input_file = open(self.cache_file, 'r')

        this_timestamp = None
        degenerate_keys = 0
        for line in input_file:
            if line.startswith('BEGIN'):
                if degenerate_keys > 0:
                    warnings.warn("%d degenerate keys found for timestamp %d"
                                  % (degenerate_keys, this_timestamp))
                this_timestamp = int(line.split()[1])
                if this_timestamp in self:
                    warnings.warn("degenerate timestamp %d found" % this_timestamp)
                self[this_timestamp] = {}
                degenerate_keys = 0
            else:
                match = _REX_LFS_DF.search(line)
                if match is not None:
                    file_system, target_name = re.findall('[^-_]+', match.group(1))[0:2]

                    if file_system not in self[this_timestamp]:
                        self[this_timestamp][file_system] = {}

                    # Duplicates can happen if a file system is doubly mounted
                    if target_name in self[this_timestamp][file_system]:
                        degenerate_keys += 1

                    self[this_timestamp][file_system][target_name] = {
                        'total_kib': int(match.group(2)),
                        'used_kib': int(match.group(3)),
                        'remaining_kib': int(match.group(4)),
                        'mount_pt': match.group(6),
                        'role': match.group(7).lower(),
                        'target_index': int(match.group(8)),
                    }

        input_file.close()

    def save_cache(self, output_file=None):
        """Serialize object into a form resembling the output of ``lfs df``.

        Args:
            output_file (str): Path to a file to which the serialized output
                should be written.  If None, print to stdout.
        """

        if output_file is None:
            self._save_cache(sys.stdout)
        else:
            with open(output_file, 'w') as output:
                self._save_cache(output)

    def _save_cache(self, output):
        """Serialize object into a form resembling the output of ``lfs df``.

        Args:
            output (file): File-like object into which resulting text should be
                written.
        """
        output.write(str(self))



"""
Provide a CLI interface for :meth:`tokio.connectors.mmperfmon.Mmperfmon.to_dataframe`
and :meth:`tokio.connectors.mmperfmon.Mmperfmon.save_cache` methods.
"""

import json
import argparse
import tokio.connectors.mmperfmon

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=str, nargs="+", help="mmperfmon text output to process")
    parser.add_argument("-j", "--json", action="store_true", default=True,
                        help="return output in JSON format")
    parser.add_argument("-c", "--csv", action="store_true", help="return output in CSV format")
    parser.add_argument("-o", "--output", type=str, default=None, help="output file")
    parser.add_argument("--host", type=str, default=None, help="hostname to output")
    parser.add_argument("--metric", type=str, default=None, help="metric to output")
    args = parser.parse_args(argv)

    if args.csv:
        if (not args.host and not args.metric) or (args.host and args.metric):
            parser.error("Must specify one --host or --metric with --csv")

    # Read from a cache file
    mmperfmon = tokio.connectors.mmperfmon.Mmperfmon(args.input[0])
    if len(args.input) > 1:
        for input_filename in args.input[1:]:
            mmperfmon.load_cache(input_filename)

    # Serialize the object
    cache_file = args.output
    if cache_file is not None:
        print("Caching to %s" % cache_file)

    if args.csv:
        if cache_file is None:
            print(mmperfmon.to_dataframe(by_host=args.host, by_metric=args.metric).to_csv())
        else:
            mmperfmon.to_dataframe(by_host=args.host, by_metric=args.metric).to_csv(cache_file)
    elif args.json:
        mmperfmon.save_cache(cache_file)
    else:
        raise Exception("No output format specified")



"""
Expose several methods of :mod:`tokio.connectors.darshan` via a
command-line interface.
"""

import json
import argparse
import tokio.connectors.darshan

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser(description='parse a darshan log')
    parser.add_argument('logfile', metavar='N', type=str, help='darshan log file to parse')
    parser.add_argument('--base', action='store_true', help='darshan log field data [default]')
    parser.add_argument('--total', action='store_true', help='aggregated darshan field data')
    parser.add_argument('--perf', action='store_true', help='derived perf data')
    parser.add_argument("-o", "--output", type=str, default=None, help="output file")
    args = parser.parse_args(argv)

    darshan = tokio.connectors.darshan.Darshan(args.logfile)
    if args.total:
        darshan.darshan_parser_total()

    if args.perf:
        darshan.darshan_parser_perf()

    if args.base or (not args.perf and not args.total):
        darshan.darshan_parser_base()

    # Serialize the object
    cache_file = args.output
    if cache_file is None:
        print(json.dumps(darshan, indent=4, sort_keys=True))
    else:
        print("Caching to %s" % cache_file)
        json.dump(darshan, open(cache_file, 'w'))



"""
Generate summary metrics from an h5lmt file.  Will be eventually replaced by the
summarize_tts command-line tool.
"""

import os
import sys
import json
import time
import math
import datetime
import argparse
import warnings
import tokio.common
import tokio.connectors.hdf5
import tokio.tools.hdf5

DATASETS_TO_BIN_KEYS = {
    'datatargets/readbytes': 'ost_read',
    'datatargets/writebytes': 'ost_write',
    'dataservers/cpuload': 'oss_cpu',
    'mdservers/cpuload': 'mds_cpu',
}

BYTES_TO_GIB = 2.0**30
DATE_FMT = "%Y-%m-%d"
DATE_FMT_STR = "YYYY-MM-DDTHH:MM:SS"

def summarize_reduced_data(data):
    """
    Take a list of LMT data sets and return summaries of each relevant key
    """
    totals = {
        'n': 0,
        'sum_bytes_read': 0,
        'sum_bytes_write': 0,
        'oss_ave': -0.0,
        'oss_max': -0.0,
        'mds_ave': -0.0,
        'mds_max': -0.0,
        'missing_ost_read': 0,
        'missing_ost_write': 0,
        'missing_oss_cpu': 0,
        'missing_mds_cpu': 0,
        'num_ost_read': 0,
        'num_ost_write': 0,
        'num_oss_cpu': 0,
        'num_mds_cpu': 0,
    }

    for timestamp in sorted(data.keys()):
        datum = data[timestamp]
        points_in_bin = (datum['indexf'] - datum['index0'])
        totals['sum_bytes_read'] += datum['sum_ost_read']
        totals['sum_bytes_write'] += datum['sum_ost_write']
        totals['n'] += points_in_bin

        if 'ave_oss_cpu' in datum:
            totals['oss_ave'] += datum['ave_oss_cpu'] * points_in_bin
        if 'ave_mds_cpu' in datum:
            totals['mds_ave'] += datum['ave_mds_cpu'] * points_in_bin

        if 'max_oss_cpu' in datum:
            if datum['max_oss_cpu'] > totals['oss_max']:
                totals['oss_max'] = datum['max_oss_cpu']
            else:
                totals['oss_max'] = totals['oss_max']

        if 'max_mds_cpu' in datum:
            if datum['max_mds_cpu'] > totals['mds_max']:
                totals['mds_max'] = datum['max_mds_cpu']
            else:
                totals['mds_max'] = totals['mds_max']

        totals['missing_ost_read'] += datum.get('missing_ost_read', 0)
        totals['missing_ost_write'] += datum.get('missing_ost_write', 0)
        totals['missing_oss_cpu'] += datum.get('missing_oss_cpu', 0)
        totals['missing_mds_cpu'] += datum.get('missing_mds_cpu', 0)
        totals['num_ost_read'] += datum.get('num_ost_read', 0)
        totals['num_ost_write'] += datum.get('num_ost_write', 0)
        totals['num_oss_cpu'] += datum.get('num_oss_cpu', 0)
        totals['num_mds_cpu'] += datum.get('num_mds_cpu', 0)

    # Derived values
    totals['ave_bytes_read_per_dt'] = totals['sum_bytes_read'] / totals['n']
    totals['ave_bytes_write_per_dt'] = totals['sum_bytes_write'] / totals['n']
    totals['sum_kibs_read'] = totals['sum_bytes_read'] / 1024.0
    totals['sum_kibs_write'] = totals['sum_bytes_write'] / 1024.0
    totals['sum_mibs_read'] = totals['sum_kibs_read'] / 1024.0
    totals['sum_mibs_write'] = totals['sum_kibs_write'] / 1024.0
    totals['sum_gibs_read'] = totals['sum_mibs_read'] / 1024.0
    totals['sum_gibs_write'] = totals['sum_mibs_write'] / 1024.0
    totals['sum_tibs_read'] = totals['sum_gibs_read'] / 1024.0
    totals['sum_tibs_write'] = totals['sum_gibs_write'] / 1024.0
    totals['ave_gibs_read_per_dt'] = totals['sum_gibs_read'] / totals['n']
    totals['ave_gibs_write_per_dt'] = totals['sum_gibs_write'] / totals['n']
    totals['oss_ave'] = totals['oss_ave'] / totals['n']
    totals['mds_ave'] = totals['mds_ave'] / totals['n']

    # Missing fractions
    for base in ['ost_read', 'ost_write', 'oss_cpu', 'mds_cpu']:
        if totals['num_%s' % base] > 0:
            totals['frac_missing_%s' % base] = float(totals['missing_%s' % base]) \
                                               / totals['num_%s' % base]

    # For convenience
    totals['ave_tibs_read_per_dt'] = totals['sum_gibs_read'] / totals['n']
    totals['ave_tibs_write_per_dt'] = totals['sum_gibs_write'] / totals['n']

    return totals

def bin_datasets(hdf5_file, dataset_names, orient='columns', num_bins=24):
    """Group many timeseries datasets into bins

    Takes a TOKIO HDF file and converts it into bins of reduced data (e.g., bin
    by hourly totals)

    Args:
        hdf5_file (connectors.Hdf5): HDF5 file from where data should be retrieved
        dataset_names (list of str): dataset names to be aggregated
        columns (str): either 'columns' or 'index'; same semantic meaning as
            pandas.DataFrame.from_dict
        num_binds (int): number of bins to generate per day

    Returns:
        Dictionary of lists.  Keys are metrics, and values (lists) are the
        aggregated value of that metric in a single timestep bin.  For example::

            {
                "sum_some_metric":      [  0,   2,   3,   1],
                "sum_someother_metric": [9.9, 2.3, 5.1, 0.2],
            }

    """
    bins_by_columns = {}
    bins_by_index = {}
    found_metrics = {}
    for dataset_key in dataset_names:
        # binned_dataset is a list of dictionaries.  Each list element
        # corresponds to a single time bin, and its key-value pairs are
        # individual metrics aggregated over that bin.
        #
        # binned_dataset = {
        #     "ave_mds_cpu": [
        #         0.86498,
        #         0.80356,
        #         ...
        #     ],
        #     "ave_oss_cpu": [
        #         39.7725,
        #         39.2170,
        #         ...
        #     ],
        #     ...
        # }

        binned_dataset = bin_dataset(hdf5_file, dataset_key, num_bins)

        # loop over time bins
        for index, counters in enumerate(binned_dataset):
            # loop over aggregate metrics in a single bin
            #   key = ave_mds_cpu
            #   value = { 0.86498, 0.80356 }
            for key, value in counters.items():
                # look for keys that appear in multiple datasets.  only record
                # values for the first dataset in which the key is encountered
                if key in found_metrics:
                    if found_metrics[key] != dataset_key:
                        if bins_by_columns[key][index] != value:
                            raise Exception("Not all timesteps are the same in HDF5")
                        continue
                else:
                    found_metrics[key] = dataset_key

                # append the value to its metric counter in bins_by_columns
                if key in bins_by_columns:
                    bins_by_columns[key].append(value)
                else:
                    bins_by_columns[key] = [value]

            # append the value to its metric counter in bins_by_index
            if 'tstart' in counters:
                index_key = int(time.mktime(counters['tstart'].timetuple()))
                if index_key not in bins_by_index:
                    bins_by_index[index_key] = {}
                bins_by_index[index_key].update(counters)

    if orient == 'columns':
        return bins_by_columns
    elif orient == 'index':
        return bins_by_index
    else:
        raise ValueError('only recognize index or columns for orient')

def bin_dataset(hdf5_file, dataset_name, num_bins):
    """Group timeseries dataset into bins

    Args:
        dataset (h5py.Dataset): dataset to be binned up

    Returns:
        list of dictionaries corresponding to bins.  Each dictionary contains
        data summarized over that bin's time interval.
    """

    base_key = DATASETS_TO_BIN_KEYS.get(dataset_name.lstrip('/'))
    if not base_key:
        raise KeyError("Cannot bin unknown dataset %s" % dataset_name)
    missing_key = "missing_" + base_key
    total_key = "num_" + base_key

    dataset = hdf5_file.get(dataset_name)
    if dataset is None:
        return []

    timestamps = hdf5_file.get_timestamps(dataset_name)[...]
    columns = hdf5_file.get_columns(dataset_name)

    if hdf5_file.schema:
        if not (timestamps.shape[0] % num_bins) == 0:
            warnings.warn("Bin count %d does not evenly divide into FSStepsDataSet size %d" %
                          (num_bins, (timestamps.shape[0])))
        dt_per_bin = int(timestamps.shape[0] / num_bins)
    else:
        # no schema means h5lmt file, which has + 1 extra timestep
        if not ((timestamps.shape[0] - 1) % num_bins) == 0:
            warnings.warn("Bin count %d does not evenly divide into FSStepsDataSet size %d" %
                          (num_bins, (timestamps.shape[0] - 1)))
        dt_per_bin = int((timestamps.shape[0] - 1) / num_bins)

    # we generate the missing data matrix for each dataset only once--it's very
    # expensive to do this multiple times
    missing_dataset = hdf5_file.get_missing(dataset_name)

    # create a list of dictionaries, where each list element is a bin
    binned_data = []
    for bin_num in range(num_bins):
        index0 = bin_num * dt_per_bin
        indexf = (bin_num+1) * dt_per_bin

        bin_datum = {
            "index0": index0,
            "indexf": indexf,
            "tstart": datetime.datetime.fromtimestamp(timestamps[index0]),
            'tend': datetime.datetime.fromtimestamp(timestamps[indexf - 1]),
        }

        bin_datum["max_" + base_key] = dataset[index0:indexf, :].max()
        bin_datum["min_" + base_key] = dataset[index0:indexf, :].min()
        bin_datum["sum_" + base_key] = dataset[index0:indexf, :].sum()
        bin_datum["ave_" + base_key] = bin_datum["sum_" + base_key] / float(indexf - index0)
        bin_datum["ave_" + base_key] /= columns.shape[0]

        bin_datum[missing_key] = missing_dataset[index0:indexf, :].sum()

        bin_datum[total_key] = (indexf - index0) * columns.shape[0]
        if bin_datum[total_key]:
            bin_datum["frac_" + missing_key] = float(bin_datum[missing_key]) / bin_datum[total_key]
        else:
            bin_datum["frac_" + missing_key] = 1.0

        if base_key.startswith('ost_'):
            for agg_key in 'max', 'min', 'ave', 'sum':
                root = "%s_%s" % (agg_key, base_key)
                bin_datum['%s_bytes' % root] = bin_datum[root]
                bin_datum['%s_kibs' % root] = bin_datum[root] / 1024.0
                bin_datum['%s_mibs' % root] = bin_datum['%s_kibs' % root] / 1024.0
                bin_datum['%s_gibs' % root] = bin_datum['%s_mibs' % root] / 1024.0
                bin_datum['%s_tibs' % root] = bin_datum['%s_gibs' % root] / 1024.0

        binned_data.append(bin_datum)

    return binned_data

def print_datum(datum=None, units='GiB'):
    """
    Print out the relevant fields from a bin containing aggregated values

    Args:
        datum (dict): the results of bin_datum
        units (str): units to print (KiB, MiB, GiB, TiB)
    """

    if datum is None:
        return "%28s %12s %12s %5s %5s %5s %5s %5s\n" % ("date/time",
                                                         "%ss read" % units.lower(),
                                                         "%ss writ" % units.lower(),
                                                         "ossav", "ossmx", "mdsav",
                                                         "mdsmx", "mssng")

    # The per-line "mssng" only counts read/write bytes, not CPU data
    frac_missing = float(datum.get('missing_ost_read', 0) + datum.get('missing_ost_write', 0)) \
                   / (datum.get('num_ost_read', 0) + datum.get('num_ost_write', 0))

    print_str = "%19s-%8s " % (datum['tstart'].strftime("%Y-%m-%d %H:%M:%S"),
                               datum['tend'].strftime("%H:%M:%S"))
    print_str += ("%%(sum_ost_read_%ss)12.2f " % units.lower()) % datum
    print_str += ("%%(sum_ost_write_%ss)12.2f " % units.lower()) % datum

    # The following data may not always be present; leave them blank if missing
    if 'ave_oss_cpu' in datum:
        print_str += "%(ave_oss_cpu)5.1f " % datum
    else:
        print_str += "%5s " % ''
    if 'max_oss_cpu' in datum:
        print_str += "%(max_oss_cpu)5.1f " % datum
    else:
        print_str += "%5s " % ''
    if 'ave_mds_cpu' in datum:
        print_str += "%(ave_mds_cpu)5.1f " % datum
    else:
        print_str += "%5s " % ''
    if 'max_mds_cpu' in datum:
        print_str += "%(max_mds_cpu)5.1f" % datum
    else:
        print_str += "%5s" % ''

    print_str += "%6.1f" % (100.0 * frac_missing)

    return print_str + "\n"

def print_data_summary(data, units='TiB'):
    """
    Print the output of the summarize_reduced_data function in a human-readable format
    """
    totals = summarize_reduced_data(data)

    print_str = ""

    # Data target summary
    print_str += ("Total read:  %%(sum_%ss_read)14.2f %s" % (units.lower(), units)) % totals
    if 'frac_missing_ost_read' in totals:
        print_str += ", %5.1f%% missing\n" % (100.0 * totals['frac_missing_ost_read'])
    else:
        print_str += "\n"
    print_str += ("Total write: %%(sum_%ss_write)14.2f %s" % (units.lower(), units)) % totals
    if 'frac_missing_ost_write' in totals:
        print_str += ", %5.1f%% missing\n" % (100.0 * totals['frac_missing_ost_write'])
    else:
        print_str += "\n"

    # Data server summary
    if math.copysign(1, totals.get('oss_max', -0.0)) > 0.0:
        print_str += "Average OSS CPU:        %(oss_ave)6.2f%%" % totals
        if 'frac_missing_oss_cpu' in totals:
            print_str += ", %5.1f%% missing\n" % (100.0 * totals['frac_missing_oss_cpu'])
        else:
            print_str += "\n"
        print_str += "Max OSS CPU:            %(oss_max)6.2f%%\n" % totals

    # Metadata server summary
    if math.copysign(1, totals.get('mds_max', -0.0)) > 0.0:
        print_str += "Average MDS CPU:        %(mds_ave)6.2f%%" % totals
        if 'frac_missing_mds_cpu' in totals:
            print_str += ", %5.1f%% missing\n" % (100.0 * totals['frac_missing_mds_cpu'])
        else:
            print_str += "\n"
        print_str += "Max MDS CPU:            %(mds_max)6.2f%%\n" % totals

    return print_str

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser(
        description='aggregate bytes in/out from h5lmt file every hour')
    parser.add_argument('h5lmt', type=str, nargs='+',
                        help='h5lmt file or basename to examine')
    parser.add_argument('--summary', action='store_true', help='print a summary of all output')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--bytes', action='store_true', help='print bytes')
    group.add_argument('--kibs', action='store_true', help='print in units of KiBs')
    group.add_argument('--mibs', action='store_true', help='print in units of MiBs')
    group.add_argument('--gibs', action='store_true',
                       help='print in units of GiBs (default)')
    group.add_argument('--tibs', action='store_true', help='print in units of TiBs')
    parser.add_argument('--bins', type=int, default=24,
                        help="number of bins per day")
    parser.add_argument('--start', type=str, default=None,
                        help="date/time to start in %s format" % DATE_FMT_STR)
    parser.add_argument('--end', type=str, default=None,
                        help="date/time to end in %s format" % DATE_FMT_STR)
    parser.add_argument('--json', action='store_true',
                        help='return json output')
    args = parser.parse_args(argv)

    if args.bytes:
        units = 'byte'
    elif args.kibs:
        units = 'KiB'
    elif args.mibs:
        units = 'MiB'
    elif args.tibs:
        units = 'TiB'
    else: # default units
        units = 'GiB'

    # Figure out the list of HDF5 files to open
    if args.start and args.end:
        hdf5_basename = os.path.basename(args.h5lmt[0])
        if len(args.h5lmt) > 1:
            warnings.warn("multiple h5lmt files specified with --start/--end; only using "
                          + hdf5_basename)
        hdf5_filenames = [x[0] for x in tokio.tools.hdf5.get_files_and_indices(
            hdf5_basename,
            list(DATASETS_TO_BIN_KEYS.keys())[0],
            datetime.datetime.strptime(args.start, DATE_FMT),
            datetime.datetime.strptime(args.end, DATE_FMT))]
    else:
        hdf5_filenames = args.h5lmt

    # Process each HDF5 file and append its bins to a global dictionary
    all_binned_data = {}
    if not args.json:
        sys.stdout.write(print_datum(None, units=units))
    for hdf5_filename in hdf5_filenames:
        hdf5_file = tokio.connectors.hdf5.Hdf5(hdf5_filename, 'r')
        binned_data = bin_datasets(hdf5_file=hdf5_file,
                                   dataset_names=list(DATASETS_TO_BIN_KEYS.keys()),
                                   orient='index',
                                   num_bins=args.bins)

        if not args.json:
            for timestamp in sorted(binned_data.keys()):
                print(print_datum(binned_data[timestamp], units=units).strip())

        all_binned_data.update(binned_data)

    # Print binned results in the desired format
    if args.json:
        if args.summary:
            to_serialize = {
                'bins': all_binned_data,
                'summary': summarize_reduced_data(all_binned_data),
            }
        else:
            to_serialize = {'bins': all_binned_data}
        print(json.dumps(to_serialize,
                         sort_keys=True,
                         indent=4,
                         cls=tokio.common.JSONEncoder))
    elif args.summary:
        sys.stdout.write(print_data_summary(all_binned_data, units=units))



#!/usr/bin/env python
"""Retrieve data stored in Elasticsearch

This module provides a wrapper around the Elasticsearch connection handler and
methods to query, scroll, and process pages of scrolling data.
"""

import copy
import time
import json
import mimetypes
import gzip
import warnings
import pandas
from .. import debug
try:
    import elasticsearch
    HAVE_ES_PKG = False
except ImportError:
    HAVE_ES_PKG = True

BASE_QUERY = {
    "query": {
        "constant_score": {
            "filter": {
                "bool": {
                    "must": [
                        {
                            "range": {
                                "@timestamp": {}
                            }
                        }
                    ]
                }
            }
        }
    }
}

class EsConnection(object):
    """Elasticsearch connection handler.

    Wrapper around an Elasticsearch connection context that provides simpler
    scrolling functionality for very long documents and callback functions to be
    run after each page is retrieved.
    """
    def __init__(self, host, port, index=None, scroll_size='1m', page_size=10000, timeout=30):
        """Configure and connect to an Elasticsearch endpoint.

        Args:
            host (str): hostname for the Elasticsearch REST endpoint
            port (int): port where Elasticsearch REST endpoint is bound
            index (str): name of index against which queries will be issued
            scroll_size (str): how long to keep the scroll search context open
                (e.g., "1m" for 1 minute)
            page_size (int): how many documents should be returned per scrolled
                page (e.g., 10000 for 10k docs per scroll)
            timeout (int): how many seconds to wait for a response from
                Elasticsearch before the query should time out

        Attributes:
            client: Elasticsearch connection handler
            page (dict): last page retrieved by a query
            scroll_pages (list): dictionary of pages retrieved by query
            index (str): name of index against which queries will be issued
            connect_host (str): hostname for Elasticsearch REST endpoint
            connect_port (int): port where Elasticsearch REST endpoint is bound
            connect_timeout (int): seconds before query should time out
            page_size (int): max number of documents returned per page
            scroll_size (int): duration to keep scroll search context open
            scroll_id: identifier for the scroll search context currently in use
            sort_by (str): field by which Elasticsearch should sort results
                before returning them as query results
            fake_pages (list): A list of ``page`` structures that should be
                returned by self.scroll() when the elasticsearch module is not
                actually available.  Used only for debugging.
            local_mode (bool): If True, retrieve query results from
                self.fake_pages instead of attempting to contact an
                Elasticsearch server
        """
        # retain state of Elasticsearch client
        self.client = None
        self.page = None
        self.scroll_pages = []
        self.index = index
        # connection info
        self.connect_host = host
        self.connect_port = port
        self.connect_timeout = timeout
        # for the scroll API
        self.page_size = page_size
        self.scroll_size = scroll_size
        self.scroll_id = None
        # for query_and_scroll
        self._num_flushes = 0
        self._filter_function = None
        self._flush_every = None
        self._flush_function = None
        self._total_hits = 0
        self._hits_since_flush = 0
        # hidden parameters to refine how Elasticsearch queries are issued
        self.sort_by = ''
        # for debugging
        self.fake_pages = []
        # if elasticsearch package is not available, we MUST run in local mode.
        # But this can also be changed at runtime if, e.g., elasticsearch _is_
        # available, but a connection cannot be made to the Elasticsearch
        # service.
        self.local_mode = HAVE_ES_PKG

        if host and port:
            self.connect()

    @classmethod
    def from_cache(cls, cache_file):
        """Initializes an EsConnection object from a cache file.

        This path is designed to be used for testing.

        Args:
            cache_file (str): Path to the JSON formatted list of pages
        """
        _, encoding = mimetypes.guess_type(cache_file)
        if encoding == 'gzip':
            input_fp = gzip.open(cache_file, 'rt')
        else:
            input_fp = open(cache_file, 'r')

        # convert cached hits into something resembling a real Elasticsearch response
        pages = []
        for hits in json.load(input_fp):
            pages.append({
                '_scroll_id': '0',
                'hits': {
                    'hits': hits
                }
            })

        # never forget to terminate fake pages with an empty page
        pages.append({'_scroll_id': 0, 'hits': {'hits': []}})
        input_fp.close()

        instance = cls(host=None, port=None, index=None)
        instance.local_mode = True
        instance.fake_pages = pages
        return instance

    def save_cache(self, output_file=None):
        """Persist the response of the last query to a file

        This is a little different from other connectors' save_cache() methods
        in that it only saves down the state of the last query's results.  It
        does not save any connection information and does not restore the state
        of a previous EsConnection object.

        Its principal intention is to be used with testing.

        Args:
            output_file (str or None): Path to file to which json should be
                written.  If None, write to stdout.  Default is None.
        """
        # write out pages to a file
        if output_file is None:
            print(json.dumps(self.scroll_pages, indent=4))
        else:
            _, encoding = mimetypes.guess_type(output_file)
            output_file = gzip.open(output_file, 'w') if encoding == 'gzip' else open(output_file, 'w')
            json.dump(self.scroll_pages, output_file)
            output_file.close()

    def _process_page(self):
        """Remove a page from the incoming queue and append it

        Takes the last received page (self.page), updates the internal state of
        the scroll operation, updates some internal counters, calls the flush
        function if applicable, and applies the filter function.  Then appends
        the results to self.scroll_pages.

        Returns:
            bool: True if hits were appended or not
        """
        if not self.page['hits']['hits']:
            return False

        self.scroll_id = self.page.get('_scroll_id')
        num_hits = len(self.page['hits']['hits'])
        self._total_hits += num_hits

        # if this page will push us over flush_every, flush it first
        if self._flush_function is not None \
        and self._flush_every \
        and (self._hits_since_flush + num_hits) > self._flush_every:
            self._flush_function(self)
            self._hits_since_flush = 0
            self._num_flushes += 1

        # increment hits since flush only after we've (possibly) flushed
        self._hits_since_flush += num_hits

        # if a filter function exists, use its output as the page to append
        if self._filter_function is None:
            filtered_page = self.page
        else:
            filtered_page = self._filter_function(self.page)

        # finally append the page
        self.scroll_pages.append(filtered_page)
        return True


    def _pop_fake_page(self):
        if not self.fake_pages:
            warn_str = "fake_pages is empty on a query/scroll; this means either"
            warn_str += "\n\n"
            warn_str += "1. You forgot to set self.fake_pages before issuing the query, or\n"
            warn_str += "2. You forgot to terminate self.fake_pages with an empty page"
            warnings.warn(warn_str)
        self.page = self.fake_pages.pop(0)

    def connect(self):
        """Instantiate a connection and retain the connection context.
        """
        if self.local_mode:
            warnings.warn("operating in local mode; elasticsearch queries will not be issued")
        else:
            self.client = elasticsearch.Elasticsearch(host=self.connect_host,
                                                      port=self.connect_port,
                                                      timeout=self.connect_timeout)
    def close(self):
        """Close and invalidate the connection context.
        """
        if self.client:
            self.client = None

    def query(self, query):
        """Issue an Elasticsearch query.

        Issues a query and returns the resulting page.  If the query included a
        scrolling request, the `scroll_id` attribute is set so that scrolling
        can continue.

        Args:
            query (dict): Dictionary representing the query to issue

        Returns:
            dict: The page resulting from the issued query.
        """
        if self.local_mode:
            self._pop_fake_page()
        else:
            if not self.client:
                # allow lazy connect
                self.connect()
            self.page = self.client.search(body=query, index=self.index, sort=self.sort_by)

        self.scroll_id = self.page.get('_scroll_id')

        return self.page

    def scroll(self):
        """Request the next page of results.

        Requests the next page of results for a query that is scrolling.
        This can only be performed if the `scroll_id` attribute is set (e.g.,
        by the ``query()`` method).

        Returns:
            dict: The next page in the scrolling context.
        """
        if self.scroll_id is None:
            raise Exception('no scroll id')

        if self.local_mode:
            self._pop_fake_page()
        else:
            if not self.client:
                # allow lazy connect
                self.connect()
            self.page = self.client.scroll(scroll_id=self.scroll_id, scroll=self.scroll_size)

        return self.page

    def query_and_scroll(self, query, source_filter=True, filter_function=None,
                         flush_every=None, flush_function=None):
        """Issue a query and retain all results.

        Issues a query and scrolls through every resulting page, optionally
        applying in situ logic for filtering and flushing.  All resulting pages
        are appended to the ``scroll_pages`` attribute of this object.

        The ``scroll_pages`` attribute must be wiped by whatever is consuming it;
        if this does not happen, `query_and_scroll()` will continue appending
        results to the results of previous queries.

        Args:
            query (dict): Dictionary representing the query to issue
            source_filter (bool or list): Return all fields contained in each
                document's _source field if True; otherwise, only return source
                fields contained in the provided list of str.
            filter_function (function, optional): Function to call before each
                set of results is appended to the ``scroll_pages`` attribute; if
                specified, return value of this function is what is appended.
            flush_every (int or None): trigger the flush function once the
                number of docs contained across all ``scroll_pages`` reaches
                this value.  If None, do not apply `flush_function`.
            flush_function (function, optional): function to call when
                `flush_every` docs are retrieved.
        """
        ### Print query
        debug.debug_print(json.dumps(query, indent=4))

        ### Run query
        time0 = time.time()

        # initialize the scroll state
        self.scroll_pages = []
        self._filter_function = filter_function
        self._flush_every = flush_every
        self._flush_function = flush_function
        self._total_hits = 0
        self._hits_since_flush = 0

        # Get first set of results and a scroll id
        if self.local_mode:
            self._pop_fake_page()
        else:
            if not self.client:
                # allow lazy connect
                self.connect()
            self.page = self.client.search(
                index=self.index,
                body=query,
                scroll=self.scroll_size,
                size=self.page_size,
                _source=source_filter,
            )

        more = self._process_page()

        # Get remaining pages
        while more:
            self.page = self.scroll()
            more = self._process_page()
        debug.debug_print("Elasticsearch query took %s seconds" % (time.time() - time0))

    def query_timeseries(self, query_template, start, end, source_filter=True,
                         filter_function=None, flush_every=None,
                         flush_function=None):
        """Craft and issue query bounded by time

        Args:
            query_template (dict): a query object containing at least one
                ``@timestamp`` field
            start (datetime.datetime): lower bound for query (inclusive)
            end (datetime.datetime): upper bound for query (exclusive)
            source_filter (bool or list): Return all fields contained in each
                document's _source field if True; otherwise, only return source
                fields contained in the provided list of str.
            filter_function (function, optional): Function to call before each
                set of results is appended to the ``scroll_pages`` attribute; if
                specified, return value of this function is what is appended.
            flush_every (int or None): trigger the flush function once the
                number of docs contained across all ``scroll_pages`` reaches
                this value.  If None, do not apply `flush_function`.
            flush_function (function, optional): function to call when
                `flush_every` docs are retrieved.
        """
        query = build_timeseries_query(query_template, start, end)

        self.query_and_scroll(
            query=query,
            source_filter=source_filter,
            filter_function=filter_function,
            flush_every=flush_every,
            flush_function=flush_function)

    def to_dataframe(self, fields):
        """Converts self.scroll_pages to CSV

        Returns:
            str: Contents of the last query's pages in CSV format
        """
        to_df = []
        for page in self.scroll_pages:
            for record in page:
                record_dict = {}
                for field in fields:
                    record_dict[field] = record['_source'].get(field)
                to_df.append(record_dict)
        return pandas.DataFrame(to_df)

def build_timeseries_query(orig_query, start, end, start_key='@timestamp', end_key=None):
    """Create a query object with time ranges bounded.

    Given a query dict and a start/end datetime object, return a new query
    object with the correct time ranges bounded.  Relies on `orig_query`
    containing at least one ``@timestamp`` field to indicate where the time
    ranges should be inserted.

    If orig_query is querying records that contain both a "start" and "end"
    time (e.g., a job) rather than a discrete point in time (e.g., a sampled
    metric), ``start_key`` and ``end_key`` can be used to modify the query to
    return all records that overlapped with the interval specified by
    ``start_time`` and ``end_time``.

    Args:
        orig_query (dict): A query object containing at least one ``@timestamp``
            field.
        start (datetime.datetime): lower bound for query (inclusive)
        end (datetime.datetime): upper bound for query (exclusive)
        start_key (str): The key containing a timestamp against which a time
            range query should be applied.
        end_key (str): The key containing a timestamp against which the upper
            bound of the time range should be applied.  If None, treat
            ``start_key`` as a single point in time rather than the start of
            a recorded process.

    Returns:
        dict: A query object with all instances of ``@timestamp`` bounded by
        `start` and `end`.
    """
    def map_item(obj, target_key, map_function, **kwargs):
        """
        Recursively walk a hierarchy of dicts and lists, searching for a
        matching key.  For each match found, apply map_function to that key's
        value.
        """
        if isinstance(obj, list):
            iterator = enumerate
            if target_key in obj:
                map_function(obj[target_key], **kwargs)
                return
        elif isinstance(obj, dict):
            iterator = dict.items
            if target_key in obj:
                map_function(obj[target_key], **kwargs)
                return
        else:
            # hit a dead end without a match
            return

        # if this isn't a dead end, search down each iterable element
        for _, value in iterator(obj):
            if isinstance(value, (list, dict)):
                # dive down any discovered rabbit holes
                map_item(value, target_key, map_function, **kwargs)
        return

    def set_time_range(time_range_obj, start_time, end_time, time_format="epoch_second"):
        """Set the upper and lower bounds of a time range
        """
        time_range_obj['gte'] = int(time.mktime(start_time.timetuple()))
        time_range_obj['lt'] = int(time.mktime(end_time.timetuple()))
        time_range_obj['format'] = time_format
        remaps[0] += 1

    def set_time(time_range_obj, operator, time_val, time_format="epoch_second"):
        """Set a single time filter
        """
        time_range_obj[operator] = int(time.mktime(time_val.timetuple()))
        time_range_obj['format'] = time_format
        remaps[0] += 1

    # note we use a single-element list so that remaps becomes mutable
    remaps = [0]

    query = copy.deepcopy(orig_query)

    if end_key is None:
        map_item(query, start_key, set_time_range, start_time=start, end_time=end)
    else:
        map_item(query, start_key, set_time, operator='lt', time_val=end)
        map_item(query, end_key, set_time, operator='gte', time_val=start)

    if not remaps[0]:
        raise RuntimeError("unable to locate timestamp in query")

    return query

def mutate_query(mutable_query, field, value, term="term"):
    """Inserts a new condition into a query object

    See https://www.elastic.co/guide/en/elasticsearch/reference/current/term-level-queries.html
    for complete documentation.

    Args:
        mutable_query (dict): a query object to be modified
        field (str): the field to which a term query will be applied
        value: the value to match for the term query
        term (str): one of the following: term, terms, terms_set, range, exists,
            prefix, wildcard, regexp, fuzzy, type, ids.

    Returns:
        Nothing.  ``mutable_query`` is updated in place.
    """
    mutable_query['query']['constant_score']['filter']['bool']['must'].append({
        term: {field: value}
    })



#!/usr/bin/env python
"""
Connector interfaces for pytokio.  Each connector provides a Python interface
into one component-level monitoring tool.
"""

### The following modules are used internally and do not need to clutter up the namespace
# cachingdb
# common

### The following modules are specific to NERSC
# nersc_jobsdb
# collectd_es
# nersc_isdct
# nersc_lfsstate

try:
    import tokio.connectors.craysdb
except ImportError:
    pass

try:
    import tokio.connectors.darshan
except ImportError:
    pass

try:
    import tokio.connectors.hdf5
except ImportError:
    pass

try:
    import tokio.connectors.hpss
except ImportError:
    pass

try:
    import tokio.connectors.lfshealth
except ImportError:
    pass

try:
    import tokio.connectors.lmtdb
except ImportError:
    pass

try:
    import tokio.connectors.slurm
except ImportError:
    pass

try:
    import tokio.connectors.esnet_snmp
except ImportError:
    pass

try:
    import tokio.connectors.mmperfmon
except ImportError:
    pass
